class ApplicationController < ActionController::Base
  # Prevent CSRF attacks by raising an exception.
  # For APIs, you may want to use :null_session instead.
  protect_from_forgery with: :exception
  before_action :sw_version,:backup_check,:get_sys_time,:activate_license
  include Util
  include BackupRestore
  include Permissions

  class NotAuthorized < StandardError; end
  class CustomError < StandardError; end

  rescue_from NotAuthorized do |exception|
    if logged_in?
      respond_to do |format|
        format.html { render 'not_authorized', :status => 403 }
        format.all { render :text => 'You are not authorized to access this section.', :status => 403 }
      end
    else
      error_msg = 'You must be logged in to access this section.'
      respond_to do |format|
        format.html { flash.now[:error] = error_msg; redirect_to log_in_url }
        format.all { render :text => error_msg, :status => 401 }
      end
    end
  end

  rescue_from CustomError do |exception|
    respond_to do |format|
      format.html { render 'backup', :status => 403 }
      format.all { render :text => 'Backup or Restore operation is performing.', :status => 403 }
    end
  end

  helper_method :current_user, :logged_in?, :can?, :get_can
  
  def activate_license
    if request.original_url.match /preface/ or \
      request.original_url.match /upgrade/ or \
      request.original_url.match /remote_assistance/ or \
      request.original_url.match /log/ or \
      request.original_url.match /system\?tab=upgrade/ or \
      request.original_url.match /system\?tab=syssetting/ or \
      request.original_url.match /system\?tab=sysoperation/ or \
      request.original_url.match /system\/reboot/ or \
      request.original_url.match /system\/poweroff/ or \
      request.original_url.match /log_in/ or \
      request.original_url.match /log_out/ or \
      request.original_url.match /change_password/ or \
      request.original_url.match /forget_password/ or \
      request.original_url.match /\/$/
      return
    end

    if request.original_url.match /system/
      if params[:tab] == 'syssetting'
        return
      end
    end
    http = create_api_conn()
    res = query_resource(http, 'activateReport', nil)
    if res[:result].length == 1
      status = res[:result][0]["status"]
      ceip_enabled = res[:result][0]["ceip_enabled"] 
      if ceip_enabled == 0
        return
      end
      if status != "activated"
          redirect_to "/"
      end
    end
  end

  def current_user
    session[:username]
  end

  def logged_in?
    current_user != nil
  end

  def sw_version
    File.open("/opt/genieacs/genieacs_version", "r") do |f|
      f.each_line do |line|
        @version = "[Version: #{line} | Feature Set: #{TpConfig::get_vers_rename(TpConfig::get_vers())}]"
      end
    end
  end

  def get_sys_time
    @server_time = Time.now.strftime("%Y-%m-%d %H:%M:%S")
  end

  def backup_check
    if request.post? or request.delete?
      if is_backup_restore_running
        raise CustomError
      end
    end
  end
  def can?(action, resource)
    if not block_given?
      is_authorized(action, resource, get_permissions)
    else
      if can?(action, resource)     
        res = yield
        #log(action, resource) if res and action != :read
        return res
      else 
        raise NotAuthorized
      end
    end
  end

  def normalize_resource_path(resource)
    File.expand_path(resource, '/') + '/'
  end

  def denormalize_resource_path(resource, resource_prefix)
    resource_prefix = normalize_resource_path(resource_prefix)
    prefix = suffix = 0
    if resource.start_with?(resource_prefix)
      prefix = resource_prefix.length
    end

    if resource[prefix, 1] == '/'
      prefix += 1
    end
    if resource.end_with? '/'
      suffix = 1
    end
    return resource[prefix .. (-1 - suffix)]
  end

  def get_permissions
   
    case Rails.configuration.auth_method
    when :db
      if $db_change == true
        fetch_permissions_from_db
      end

      if not Rails.configuration.respond_to?("users")
        fetch_permissions_from_db
      end

      if Rails.configuration.users.blank?
        fetch_permissions_from_db
      end
      $db_change = false
    end

    roles = ['anonymous']
    if current_user
      roles.concat(Rails.configuration.users[current_user]['roles'])
    end

    @permissions ||= Rails.cache.fetch("#{roles}_permisions", :expires_in => 60.seconds) do
      permissions = []
      roles.each do |role|
        if Rails.configuration.permissions.has_key?(role)
          # do not concat directly to avoid modifying original permissions during normalization
          Rails.configuration.permissions[role].map do |a|
            permissions << a.dup
          end
        end
      end
      normalize_permissions(permissions)
    end
  end

  def fetch_permissions_from_db
    begin
      default_privileges = {}
      custom_privileges = {}
      roles = Role.all
      Rails.configuration.permissions = {}
      roles.each do |role|
        Rails.configuration.permissions[role.name] = Array.new
          default_privileges[role.id] = {}
       
        role.default_privileges.each do |privilege|
          if !default_privileges[privilege.role_id].has_key?(privilege.resource)
            default_privileges[privilege.role_id][privilege.resource] = []
          end
          default_privileges[privilege.role_id][privilege.resource].push([privilege.action, privilege.weight, privilege.resource])
        end
      
        custom_privileges[role.id] = {}
        # for upgrade, 
        # get privilegs form default_privileges 
        # if resource is new module (action 'exist' is not existed)
        role.privileges.each do |privilege|
          if privilege.action == 'exist'
            default_privileges[role.id].except!(privilege.resource)
          end
          
          if !custom_privileges[privilege.role_id].has_key?(privilege.resource)
            custom_privileges[privilege.role_id][privilege.resource] = []
          end
          custom_privileges[privilege.role_id][privilege.resource].push([privilege.action, privilege.weight, privilege.resource])
        end

        
        custom_privileges[role.id].each do |custom_resource,custom_privileges|
          custom_privileges.each do |custom_privilege|
            Rails.configuration.permissions[role.name].push(custom_privilege)
          end
        end

        default_privileges[role.id].each do |default_resource,default_privileges|
          default_privileges.each do |default_privilege|
            Rails.configuration.permissions[role.name].push(default_privilege)
          end
        end
      end

      users = User.all
      Rails.configuration.users = {}

      users.each do |user|
        Rails.configuration.users[user.username] = Hash.new
        Rails.configuration.users[user.username]["password"] = user.password
        Rails.configuration.users[user.username]["roles"] = Array.new
        user.roles.each do |role|
          Rails.configuration.users[user.username]["roles"].push(role.name)
        end
      end
    rescue
      retry
    end
  end

  def status
    http = create_api_conn()
    res = http.get('/application/status', nil)
    if res.code == '404'
      render(json: { 'msg' => 'not found' })
    elsif res.code == '200'
      data = JSON.parse(res.body)
      render json: { 'msg' => data['status'] }
    end
  end
end
