include Parser

class DeviceEntry
    include DevicesShowLib2
    include Util

    @tab = nil
    @device_params = nil
    @device_params_groups = nil
    @lan_entry = nil
    @wan_entries = nil
    @invalid_wan_entries = nil
    @ssid_entries = nil
    # http: the http connection for nbi
    # device_id
    # tab: the tab of customized pages, including 'lan', 'wan', 'wlan'
    #      To load the related nodes by the given tab, defined in config/tplink/device_info_parameters.yml
    #      If the given tab is nil, load all nodes of the device.
    #      
    def initialize(http, device_id, tab)
        @tab = tab
        @device_params = {}
        dump_params(get_data(http, device_id, tab), @device_params)
        @device_params_groups = group_params(@device_params)
    end

    # return:
    #   the cloned instacne variable of @device_params
    #   , which is a hash table with key-value pair.
    #     - key: full path of leaf node.
    #     - value: {'value'=>'xxx', 'writable'=>'xxx', 'path'=>'xxx'} 
    def device_params()
        return @device_params.clone
    end

    # tab: 'lan', 'wan', 'wlan'
    # return:
    #    nil if the given tab is nil. 
    #    the corresponding entries of the given tab, otherwise.
    #    The structure of an entry is defined in config/tplink/device_info_parameters.yml
    def get_entries(tab = @tab)
        return nil if tab.nil?

        ret =  nil
        case tab
        when 'lan'
            ret = get_lan_entries
        when 'wan'
            ret = get_wan_entries
        when 'wlan'
            ret = get_ssid_entries
        end
        #puts ret.to_json
        return ret
    end

    # Get path from entry of wan
    # interface_name: 
    #   the name of "Interface Name" from UI
    # fields: 
    #   a array of nodes name
    # return:
    #   nil if @wan_entries is nil or unknown
    #   a array of paths
    def get_wan_entry_paths(interface_name, fields)
        initialize_wan_entries()
      
        @wan_entries.each do |wan_entry|
            if wan_entry.getValue('wan_ifc_name')['value'] == interface_name
                return wan_entry.getParameterPaths(fields)
            end
        end
        return nil
    end

    # Generate the set parameters
    # params: 
    #    a hash table with key-value pair, see the structure of get_entries
    # tab:
    #    the tab of customized page, including 'lan', 'wan' and 'wlan'
    # return:
    #    nil if the given tab is nil or unknown
    #    the set parameters if it is changed.
    def generate_set_parameters(params, tab = @tab)
        return nil if tab.nil?

        ret =  nil
        case tab
        when 'lan'
            ret = generate_lan_parameters params
        when 'wan'
            ret = generate_wan_parameters params
        when 'wlan'
            ret = generate_ssid_parameters params
        end
        return ret
    end

    # return the list of lan entry.
    #    The structure is defined in config/tplink/device_info_parameters.yml
    def get_lan_entries() 
        ret = []
        @lan_entry = LanEntry.new('Device', @device_params, @device_params_groups) if @lan_entry.nil?
        ret << @lan_entry.getParameters
        return ret;
    end

    # return set parameter of the given params
    #    The structure of an entry is defined in config/tplink/device_info_parameters.yml
    def generate_lan_parameters(params)
        ret = []
        @lan_entry = LanEntry.new('Device', @device_params, @device_params_groups) if @lan_entry.nil?
        a = @lan_entry.generate_set_parameters(params)
        a.each do |e|
            e[1].each do |path, value|
                ret << [path, value]
            end
        end
        return ret
    end

    # initialize the entries of wan.
    # The structure of an entry is defined in config/tplink/device_info_parameters.yml
    def initialize_wan_entries() 
        if @wan_entries.nil? 
            @wan_entries = []
            paths = get_wan_interface_paths
            known_ip_interfaces = []
            paths.each do |path|
                wan_entry = nil
                begin
                    wan_conn_type = get_wan_conn_type(path)
                    #puts "get_wan_conn_type(#{path}) = #{wan_conn_type}"
                    case wan_conn_type
                    when 'static'
                        wan_entry = WanStaticEntry.new('Device', @device_params, @device_params_groups, path, wan_conn_type)
                    when 'dynamic'
                        wan_entry = WanDynamicEntry.new('Device', @device_params, @device_params_groups, path, wan_conn_type)
                    when 'ppp'
                        wan_entry = WanPppoeEntry.new('Device', @device_params, @device_params_groups, path, wan_conn_type)
                    when 'l2tp'
                        wan_entry = WanL2tpEntry.new('Device', @device_params, @device_params_groups, path, wan_conn_type)
                    when 'pptp'
                        wan_entry = WanPptpEntry.new('Device', @device_params, @device_params_groups, path, wan_conn_type)
                    end
                rescue => exception 
                    wan_entry = nil
                    puts "WanEntry: Some Exception happens (path = #{path}, wan_conn_type = #{wan_conn_type}):"
                    puts exception.backtrace 
                end
                @wan_entries << wan_entry if !wan_entry.nil?
            end

            # Collect the invalid wan entries of Device.IP.Interface.\\d 
            # because Device.X_TP_FakeIntf is not readable.
            @invalid_wan_entries = get_invalid_wan_entries()
            
        end
    end

    # Collect the invalid 'Device.IP.Interface.\\d'
    # because Device.X_TP_FakeIntf is not readable.
    def get_invalid_wan_entries() 
        ret = []
        known_ip_interfaces = []
        @wan_entries.each do | wan_entry |
            #puts "#{wan_entry.ip_interface}, #{wan_entry.wan_conn_type}, #{wan_entry.ip_interface_xxtp}"
            case wan_entry.wan_conn_type
            when 'l2tp', 'pptp'
                ifc2 = wan_entry.ip_interface_xxtp
                if !ifc2.nil?
                    a, b, c  = ifc2.rpartition('.')
                    known_ip_interfaces << a
                end
            end
        end
        
        l2tps = match_rkey_rval(@device_params, "Device.IP.Interface.\\d+.LowerLayers", "Device.X_TP_L2TP.Interface.\\d+.")
        pptps = match_rkey_rval(@device_params, "Device.IP.Interface.\\d+.LowerLayers", "Device.X_TP_PPTP.Interface.\\d+.") 
        
        entries_1 = get_xxtp_entries('l2tp', l2tps, known_ip_interfaces)
        entries_2 = get_xxtp_entries('pptp', pptps, known_ip_interfaces)
        
        ret = ret.concat(entries_1).concat(entries_2)
        
        return ret
    end
    
    # Return the wan entry of invalid l2tp/pptp.
    # wan_conn_type: 
    #    'l2tp' or 'pptp'
    # entries: 
    #    list of @device_params where
    #      Device.IP.Interface.\\d+.LowerLayers = Device.X_TP_L2TP.Interface.\\d+
    #      ,or Device.IP.Interface.\\d+.LowerLayers = Device.X_TP_PPTP.Interface.\\d+
    # known_ip_interfaces
    #    known ip interfaces
    def get_xxtp_entries(wan_conn_type, entries, known_ip_interfaces) 
        ret = []
        entries.each do | entry |
            path = entry['path']
            ifc, b, c = path.rpartition('.')
            found = false 
            known_ip_interfaces.each do | known_ip_interface |
                found = true if known_ip_interface == ifc 
            end
            
            if !found 
                #puts "not found #{ifc}"
                case wan_conn_type
                when 'l2tp'
                    ret << WanL2tpEntry.new('Device', @device_params, @device_params_groups, nil, ct, "#{ifc}.")
                when 'pptp'
                    ret << WanPptpEntry.new('Device', @device_params, @device_params_groups, nil, ct, "#{ifc}.")
                end
            end
        end
        
        return ret
    end

    # Return the wan entries.
    # The structure of an entry is defined in config/tplink/device_info_parameters.yml
    def get_wan_entries() 
        ret = []
        initialize_wan_entries()
        @wan_entries.each do |wan_entry|
            ret << wan_entry.getParameters
        end
        
        @invalid_wan_entries.each do |wan_entry|
            ret << wan_entry.getParameters
        end
        
        return ret 
    end
    
    # return set parameter of the given params
    #    The structure of an entry is defined in config/tplink/device_info_parameters.yml
    def generate_wan_parameters(params)
        return nil if !params.key?('wan_ifc_name') || !params.key?('wan_conn_type') 

        #puts "---------------------"
        #puts "generate_wan_parameters: params = #{params}"
        #puts "---------------------"
        # require 'tplink/parser'
        ret = []
        initialize_wan_entries()
        @wan_entries.each do | wan_entry | 
            wan_ifc_name = wan_entry.getValue('wan_ifc_name')['value']
            wan_conn_type = wan_entry.getValue('wan_conn_type')['value']
            next if diff?(wan_ifc_name, params['wan_ifc_name']) || diff?(wan_conn_type, params['wan_conn_type'])
            
            a = wan_entry.generate_set_parameters(params)
            a.each do |e|
                e[1].each do |path, value|
                    ret << [path, value]
                end
            end
            
            break
        end
        #puts "---------------------"
        #puts "generate_wan_parameters: ret = #{ret}"
        #puts "---------------------"
        return ret
    end

    # Return the list of Device.IP.Interface.{i}
    # , which is associated with Device.Ethernet.Interface.{i}.Upstream = true
    def get_wan_interface_names
        ret = []
        paths = get_wan_interface_paths
        paths.each do | path |
            # path = "Device.IP.Interface.\\d+."
             a, b, c = path.rpartition('.')
            # a = Device.IP.Interface.\\d+
            ret << {'name'=>@device_params["#{path}Name"]['value'], 'interface_path'=> a}
        end
        return ret
    end

    def get_wan_conn_type(ip_interface) 
        return DeviceEntry::get_wan_conn_type(@device_params, ip_interface)
    end

    # return the wan connection type, including 'static', 'dynamic', 'ppp', 'l2tp' and 'pptp'
    # see config/tplink/device_subpage_parameters.yml
    def self.get_wan_conn_type(device_params, ip_interface) 
        ret = 'Unknown'
        begin
            lower_layers = device_params["#{ip_interface}LowerLayers"]

            if rmatch?(lower_layers['value'], "Device.Ethernet.VLANTermination.\\d+.") 
                dhcp_clients = match_rkey_val(device_params, "Device.DHCPv4.Client.\\d+.Interface", ip_interface)
                case dhcp_clients.size
                when 0
                    addressing_type = device_params["#{ip_interface}IPv4Address.1.AddressingType"]
                    ret = 'static' if addressing_type['value'] == 'Static'
                when 1
                    dhcp_client = dhcp_clients[0]
                    dhcp_client_en = device_params[dhcp_client['path'].sub('Interface', 'Enable')]
                    return ret if dhcp_client_en.nil? 
                    ret = 'dynamic'
                end
                
                fake_intfs = match_rkey_val(device_params, "Device.X_TP_FakeIntf.\\d+.Interface", ip_interface)
                if fake_intfs.size > 0
                    prefix, match, suffix = fake_intfs[0]['path'].rpartition('.')
                    
                    l2tps = match_rkey_val(device_params, "Device.X_TP_L2TP.Interface.\\d+.LowerLayers", "#{prefix}#{match}") 
                    return 'l2tp' if l2tps.size > 0
                    
                    pptps = match_rkey_val(device_params, "Device.X_TP_PPTP.Interface.\\d+.LowerLayers", "#{prefix}#{match}") 
                    return 'pptp' if pptps.size > 0
                end
                
            elsif rmatch?(lower_layers['value'], "Device.PPP.Interface.\\d+.") 
                return 'ppp'
            end
            
        rescue => exception
            #puts "#{exception}"
            ret = 'Unknown'
        end
        return ret
    end

    def get_wan_interface_paths
        ethernet_interfaces = []
        results = get_device_param_by_group(@device_params_groups, "Device.Ethernet.Interface.\\d+", "Upstream")
        results.each do | result |
            if result.key?('value') && result['value'].eql?("true")
                prefix, match, suffix =  result['path'].rpartition('.')
                ethernet_interfaces << prefix if @device_params["#{prefix}.Enable"]['value'] == "true"
            end
        end
        ip_interfaces = []
        ethernet_interfaces.each do | ethernet_interface |
            a = get_ip_interfaces(ethernet_interface)
            ip_interfaces = ip_interfaces.concat (a ) if !a.nil?
        end
        ip_interfaces = ip_interfaces.uniq
        return ip_interfaces
    end
    
    # Return the ip interfaces which is associated with Device.Ethernet.Interface.\\d+.Upstream = true
    def get_ip_interfaces(ethernet_interface) 
        ret = []
        arr = get_device_param_by_group(@device_params_groups, 'Device.InterfaceStack.\\d+', 'HigherLayer', 'Device.IP.Interface.\\d+.')
        
        arr.each do | higher_layer |
            next if higher_layer['value'] == 'Device.IP.Interface.1.'
            val = find_InterfaceStack(higher_layer['path'], ethernet_interface)
            ret << higher_layer['value'] if !val.nil?
        end
        return ret
    end
    
    def find_InterfaceStack(higher_layer_path, target)
        ret = nil
        
        prefix, match, suffix = higher_layer_path.rpartition('.')
        
        higher_layer_val = @device_params[higher_layer_path]['value']
        lower_layer_path = "#{prefix}.LowerLayer"
        lower_layer_val = @device_params["#{prefix}.LowerLayer"]['value']
        
        if rmatch?(lower_layer_val, target)
            return lower_layer_val
        else 
            arr = get_device_param_by_group(@device_params_groups, 'Device.InterfaceStack.\\d+', 'HigherLayer', lower_layer_val)
            
            arr.each do | a |
                prefix, match, suffix =  a['path'].rpartition('.')
                next_higher_layer_path = a['path']
                next_higher_layer_val  = a['value']
                next_lower_layer_path  = "#{prefix}.LowerLayer"
                next_lower_layer_val   = @device_params["#{prefix}.LowerLayer"]['value']
                
                ret = find_InterfaceStack(next_higher_layer_path, target)
                if ret.nil?
                    next 
                else 
                    return ret
                end
            end
        end
        
        return ret
    end

    # initialize ssid entries
    # The structure of an entry is defined in config/tplink/device_info_parameters.yml#wlan
    def initialize_ssid_entries() 
        if @ssid_entries.nil? 
            paths = get_ssid_paths
            if @ssid_entries.nil? 
                @ssid_entries = []
                paths.each do |path|
                    begin
                        ssid_entry = SsidEntry.new('Device', @device_params, @device_params_groups, path['access_point_path'], path['ssid_path'], path['radio_path'])
                    rescue => exception 
                        ssid_entry = nil
                        puts "SsidEntry: Some Exception happens (path = #{path})."
                        puts exception.backtrace 
                    end
                    @ssid_entries << ssid_entry if !ssid_entry.nil?
                end
            end
        end
    end

    # Return the ssid entries.
    # The structure of an entry is defined in config/tplink/device_info_parameters.yml#wlan
    def get_ssid_entries() 
        ret = []
        initialize_ssid_entries()
        @ssid_entries.each do |ssid_entry|
            ret << ssid_entry.getParameters
        end
        # puts "get_ssid_entries: ret = #{ret}"
        return ret 
    end

    # return set parameter of the given params
    #    The structure of an entry is defined in config/tplink/device_info_parameters.yml
    def generate_ssid_parameters(params)
        return nil if !params.key?('index')
        # require 'tplink/parser'
        ret = []
        initialize_ssid_entries()
        index = nil
        ap_type = nil
        band = nil
        band_steering = nil
        array = []
        guest_index_2g = 0
        guest_index_5g = 0
        guest_index = 0
        @ssid_entries.each do | ssid_entry | 
            index = ssid_entry.getValue('index')['value']
            guest_index_2g += 1 if ssid_entry.getValue('ap_type')['value'] == "Guest" && ssid_entry.getValue('band')['value'] == "2.4GHz"
            guest_index_5g += 1 if ssid_entry.getValue('ap_type')['value'] == "Guest" && ssid_entry.getValue('band')['value'] == "5GHz"
            next if diff?(params['index'], index)
            
            ap_type = ssid_entry.getValue('ap_type')['value']
            band = ssid_entry.getValue('band')['value']
            band_steering = ssid_entry.getValue('band_steering')['value']
            array << ssid_entry.generate_set_parameters(params)

            # set parameters for band_steering = on
            if ap_type == "Guest"
                guest_index = ( band == "5GHz") ? guest_index_5g : guest_index_2g
            end
            band = ( band == "5GHz") ? "2.4GHz" : "5GHz"
            
            if band_steering == "on"
                @ssid_entries.each do | entry |
                    next if diff?(band, entry.getValue('band')['value']) or diff?(ap_type, entry.getValue('ap_type')['value']) 
                    guest_index -=1 
                    next if ap_type == "Guest" && guest_index != 0

                    params_2 = params.clone
                    params_2['index'] = entry.getValue('index')['value']
                    params_2 = params_2.except('Standard', 'channel', 'channel_bandwidth')
                    
                    array << entry.generate_set_parameters(params_2)
                end 
            end
        end

        array.each do | element |
            element.each do |e|
                e[1].each do |path, value|
                    next if ap_type == "Guest" && path.start_with?("Device.WiFi.Radio.")
                    ret << [path, value]
                end
            end
        end
        #puts ">>>> ret = #{ret}"
        return ret
    end

    # return [
    #   { 
    #     'ssid_path' => 'Device.WiFi.SSID.1', 
    #     'access_point_path' => 'Device.WiFi.AccessPoint.1', 
    #     'radio_path' => ''Device.WiFi.Radio.1' 
    #   },
    #    ...
    # ]
    def get_ssid_paths()
        ret = []
        list = get_device_params(@device_params, "Device.WiFi.SSID.\\d+.Enable")
        
        list.each do | ssid_enable |
            ssid_path = ssid_enable['path'].sub(".Enable", "")
            access_point_path = nil
            radio_path = nil
            
            res = match_rkey_val(@device_params, "Device.WiFi.AccessPoint.\\d+.SSIDReference", "#{ssid_path}.")
            
            res.each do | ssid_reference |
                access_point_path, b, c = ssid_reference['path'].rpartition('.')
            end
            
            if !access_point_path.nil?
                ap_type_path = "#{access_point_path}.X_TP_APType"
                ap_type = get_device_param_by_specific( @device_params, ap_type_path)
                if !ap_type.nil?
                    next if ap_type['value'] != 'Primary' && ap_type['value'] != 'Guest'
                end
            end
            
            radio_refObj = get_device_param_by_specific(@device_params, "#{ssid_path}.LowerLayers")
            if !radio_refObj.nil?
                radio_path = radio_refObj['value']
                if @device_params.key? "#{radio_path}OperatingFrequencyBand"
                    a, b, c = radio_path.rpartition('.')
                    radio_path = a
                end
            end
            
            ret << { 'ssid_path' => ssid_path, 'access_point_path' => access_point_path, 'radio_path' => radio_path }
        end
        
        return ret
    end
end
