module BackhaulLinkType
    None = "None"
    ETHERNET = "Ethernet"
    WIFI = "Wi-Fi"
end


class MeshEntry
    include Parser
    @@WIFI_QUALITY = {}
    @@WIFI_QUALITY["disconnected"] = 0
    @@WIFI_QUALITY["poor"] = 1
    @@WIFI_QUALITY["normal"] = 2
    @@WIFI_QUALITY["good"] = 3

    # mac_1: {
    #    "path"=>"Device.WiFi.MultiAP.APDevice.1.MACAddress"
    #    "value"=>"AA:BB:CC:DD:EE:01", 
    #    "writable"=>"false", 
    #    "mesh_type"=>"<MeshType>", 
    #    "device_name"=>"<device_name>"
    #    [, "re_mapping_path"=>"Device.X_TP_RE_MAPPING.1" ]  # for satellite only
    # }
    # mac_2: {
    #    "path"=>"Device.WiFi.MultiAP.APDevice.1.BackhaulMACAddress", 
    #    "value"=>"", 
    #    "writable"=>"false", 
    #    "device_name"=>"<device_name>" 
    #}
    # link_type: { "path"=>"Device.WiFi.MultiAP.APDevice.1.BackhaulLinkType", "value"=>"Ethernet", "writable"=>"false"}
    def initialize(device_params, device_params_groups, mac_1, mac_2, link_type, controller)
        @device_params = device_params
        @device_params_groups = device_params_groups
        @mac_1 = mac_1
        @mac_2 = mac_2
        @link_status = 'unknown'
        @link_type = link_type
        @controller = controller
        @re_mapping_path = @mac_1['re_mapping_path']
        @re_mac_list = get_satellites_mac_list()
        @params = {}
        initValues()
    end

    def setValue(field, value)
        @params[field] = value.nil? ? "" : value 
    end

    def getValue(field)
        return @params[field].clone
    end
    
    def setValue_by_specific_path(field, specific_path, default_value) 
        value = get_device_param_by_specific(@device_params, specific_path)
        value = default_value if value.nil?

        if value.include?("value")
            value = value["value"]
        end
        setValue(field, value)
    end

    def initValues()
        call_functions([{ "name" => "init_mac_1" }])
        call_functions([{ "name" => "init_mac_2" }])
        call_functions([{ "name" => "init_link_status" }])
        call_functions([{ "name" => "init_link_type" }])
        call_functions([{ "name" => "init_dev_name_1" }])
        call_functions([{ "name" => "init_dev_name_2" }])

        case getValue('link_type')
        when BackhaulLinkType::ETHERNET
            call_functions([{ "name" => "init_ethernet_status" }])
        when BackhaulLinkType::WIFI
            call_functions([{ "name" => "init_wifi" }])
        end

        call_functions([{ "name" => "init_clients_num" }])
        # case @controller
        # when "devices"
        #     call_functions([{ "name" => "init_clients_num" }])
        # when "wifi_mgmt"
        #     call_functions([{ "name" => "init_clients" }])
        # end

    end

    def init_mac_1() 
        setValue('mac_1', @mac_1['value'])
    end

    def init_mac_2() 
        setValue('mac_2', @mac_2['value'])
    end

    def init_link_status() 
        @link_status = 'unknown'
        o = get_device_param_by_specific(@device_params, "#{@re_mapping_path}.Status.Active")
        if !o.nil? and o.include?('value') and !o['value'].blank?
            case o['value']
            when "false"
                @link_status = 'inactive'
            when "true"
                @link_status = 'active'
            else 
                @link_status = "unknown value(#{o['value']})"
            end
        end
    end

    def init_link_type() 
        if @link_type['value'].blank? or @link_type['value'].eql?("None") or @mac_2['value'].blank? or @mac_2['value'].eql?("00:00:00:00:00:00")
            setValue('link_type', BackhaulLinkType::None)
        else
            setValue('link_type', @link_type['value'])
            setValue('link_type', BackhaulLinkType::WIFI) if @link_status != "active"
        end
    end

    def init_dev_name_1()
        setValue('dev_name_1', @mac_1['device_name'])
    end

    def init_dev_name_2()
        setValue('dev_name_2', @mac_2['device_name'])
    end

    def init_ethernet_status() 
        #puts "init_ethernet_status: #{@re_mapping_path}"
        return if @re_mapping_path.nil?
        setValue('ethernet', { "status" => @link_status })
    end

    def init_wifi()
        return if @re_mapping_path.nil?
        wifi = []
        bands = {'24g' => "2.4GHz" , '5g' => "5GHz"}
        bands.each do | k, v |
            h = {}
            pp = get_wifi_parameters(k)
            pp.each do |kk, vv|
                o = get_device_param_by_specific(@device_params, vv)
                if !o.nil? and o.include?('value')
                    case kk 
                    when 'status'
                        h = init_wifi_status(kk, o['value'], h)
                    when 'link_rate'
                        h = init_wifi_link_rate(kk, o['value'], h)
                    when 'quality'
                        h = init_wifi_quality(kk, o['value'], h)
                    when 'rssi'
                        #tranfer QCA RSSI value: rssi = rssi - noise(97)
                        value = o['value'].to_i
                        if value > 0
                            value = value - 97;
                        end
                        h = init_wifi_rssi(kk, value, h)
                    else
                        h[kk] = (!o['value'].blank?) ? o['value'] : ""
                    end
                end
            end
            if !h.blank? 
                h['band'] = v 
                wifi << h
            end
        end
        setValue('wifi', wifi) if !wifi.blank?
    end

    # append_str: 24g or 5g
    def get_wifi_parameters(append_str)
        ret = {}
        ret["status"] = "#{@re_mapping_path}.Status.ConnWlMac#{append_str}"
        ret["link_rate"] = "#{@re_mapping_path}.Status.ConnWlTxRate#{append_str}"
        ret["quality"] = "#{@re_mapping_path}.Status.ConnWlQuality#{append_str}"
        ret["rssi"] = "#{@re_mapping_path}.Status.ConnWlRssi#{append_str}"
        return ret
    end

    def init_wifi_status(k, v, hash)
        return hash if k != "status"

        hash[k] = ""
        if @link_status != "active" or v.blank? or v.eql?("00:00:00:00:00:00")
            hash[k] = "inactive"
        else
            hash[k] = "active"
        end 
        return hash
    end

    def init_wifi_link_rate(k, v, hash)
        return hash if k != "link_rate"
        case hash['status']
        when 'active'
            hash[k] = "#{v} Mbit/s"
        end
        return hash
    end

    def init_wifi_quality(k, v, hash)
        return hash if k != "quality"
        case hash['status']
        when 'active'
            hash[k] = "unknown(#{v})"
            @@WIFI_QUALITY.each do | _text, _val |
                if v.to_s == _val.to_s
                    hash[k] = _text
                end
            end
        end
        return hash
    end

    def init_wifi_rssi(k, v, hash)
        return hash if k != "rssi"
        case hash['status']
        when 'active'
            hash[k] = "#{v} dBm" if !v.nil?
        end
        return hash
    end

    def init_clients_num
        case @mac_1['mesh_type']
        when MeshType::CONTROLLER
            calculate_by_mac = true
            if calculate_by_mac
                setValue('clients_num', get_clients_num(getValue('mac_1')))
            else 
                num_1 = get_clients_num() 
                num_2 = get_clients_num(@re_mac_list) 
                num = {"total" => 0, "wired" => 0, "wireless" => 0}
                if !num_1.blank?
                    num["total"] = num_1["total"]
                    num["wired"] = num_1["wired"]
                    num["wireless"] = num_1["wireless"]
                end
                if !num_2.blank?
                    num["total"] =  num["total"] - num_2["total"]
                    num["wired"] =  num["wired"] - num_2["wired"]
                    num["wireless"] =  num["wireless"] - num_2["wireless"]
                end
                setValue('clients_num', num )
            end
        when MeshType::SATELLITE
            setValue('clients_num', get_clients_num(getValue('mac_1')) )
        end
    end

    def init_clients
        info = get_clients_info(getValue('mac_1'))
        setValue('clients_info', info )
    end

    def get_satellites_mac_list()
        res = get_device_params( @device_params, "Device.WiFi.MultiAP.APDevice.\\d+.BackhaulMACAddress")
        return nil if res.nil?
        mac_list = []
        res.each do |o|
            next if o['value'].blank? or o['value'] == "00:00:00:00:00:00"
            prefix, match, subfix = o['path'].rpartition('.')
            v = get_device_param_by_specific(@device_params, "#{prefix}.BackhaulLinkType")
            next if v['value'].blank? or v['value'] == BackhaulLinkType::None

            v = get_device_param_by_specific(@device_params, "#{prefix}.MACAddress")
            if !v.blank? and !v['value'].blank?
                mac_list << v['value']
            end
        end
        return mac_list
    end

    # mac_list:
    #   nil for total client number 
    #       => COUNT_IF( 
    #              "Device.Hosts.Host.{i}.Active" = true &&
    #              NOT_RE( "Device.Hosts.Host.{i}.PhysAddress" ) )
    #   a MAC string for the client number of given mac 
    #       => COUNT_IF( 
    #              "Device.Hosts.Host.{i}.Active" = true && 
    #              NOT_RE( "Device.Hosts.Host.{i}.PhysAddress" ) &&
    #              "Device.Hosts.Host.{i}.X_TP_DevphyAddress" = <MAC> )
    #   an array of MAC for the client number of given array 
    #        => SUM(COUNT_IF( 
    #              "Device.Hosts.Host.{i}.Active" = true && 
    #              NOT_RE( "Device.Hosts.Host.{i}.PhysAddress" ) &&
    #              "Device.Hosts.Host.{i}.X_TP_DevphyAddress" = <EACH_MAC> ))
    # 
    def get_clients_num(mac_list=nil) 
        ret = {"total" => 0, "wired" => 0, "wireless" => 0}
        return ret if mac_list.is_a?(Array) and mac_list.blank?

        if mac_list.nil?
             res = match_rkey_val(@device_params, "Device.Hosts.Host.\\d+.Active", "true")
             res.each do | o |
                ret['total'] = ret['total'] + 1
                prefix, match, subfix = o['path'].rpartition('.')
                
                physAddress = get_device_param_by_specific( @device_params, "#{prefix}.PhysAddress") 
                next if !physAddress.blank? && !physAddress['value'].blank? && @re_mac_list.include?(physAddress['value'])
                
                lanConnType = get_device_param_by_specific( @device_params, "#{prefix}.X_TP_LanConnType") 
                if !lanConnType.blank?
                    case lanConnType['value']
                    when "0" #wired(0)
                        ret['wired'] = ret['wired'] + 1 
                    when "1" #wireless(1)
                        ret['wireless'] = ret['wireless'] + 1 
                    end
                end
             end
        elsif mac_list.is_a?(String)
            res = match_rkey_val(@device_params, "Device.Hosts.Host.\\d+.Active", "true")
            res.each do | o |
                prefix, match, subfix = o['path'].rpartition('.')
                
                physAddress = get_device_param_by_specific( @device_params, "#{prefix}.PhysAddress") 
                next if !physAddress.blank? && !physAddress['value'].blank? && @re_mac_list.include?(physAddress['value'])
                
                mac = get_device_param_by_specific( @device_params, "#{prefix}.X_TP_DevphyAddress")
                if !mac.blank? && !mac['value'].blank? && mac['value'] == mac_list
                    ret['total'] = ret['total'] + 1
                    lanConnType = get_device_param_by_specific( @device_params, "#{prefix}.X_TP_LanConnType") 
                    if !lanConnType.blank?
                        case lanConnType['value']
                        when "0" #wired(0)
                            ret['wired'] = ret['wired'] + 1 
                        when "1" #wireless(1)
                            ret['wireless'] = ret['wireless'] + 1 
                        end
                    end
                end
            end
        elsif mac_list.is_a?(Array)
            res = match_rkey_val(@device_params, "Device.Hosts.Host.\\d+.Active", "true")
            res.each do | o |
                prefix, match, subfix = o['path'].rpartition('.')
                
                physAddress = get_device_param_by_specific( @device_params, "#{prefix}.PhysAddress") 
                next if !physAddress.blank? && !physAddress['value'].blank? && @re_mac_list.include?(physAddress['value'])
                
                mac = get_device_param_by_specific( @device_params, "#{prefix}.X_TP_DevphyAddress")
                if !mac.blank? && !mac['value'].blank? && mac_list.include?(mac['value'])
                    ret['total'] = ret['total'] + 1
                    lanConnType = get_device_param_by_specific( @device_params, "#{prefix}.X_TP_LanConnType") 
                    if !lanConnType.blank?
                        case lanConnType['value']
                        when "0" #wired(0)
                            ret['wired'] = ret['wired'] + 1 
                        when "1" #wireless(1)
                            ret['wireless'] = ret['wireless'] + 1 
                        end
                    end
                end
            end
        end
        return ret
    end

    def get_clients_info(mac_list=nil)
        ret = {"wired"=>0,"good"=>0,"normal"=>0,"poor"=>0}
        res = match_rkey_val(@device_params, "Device.Hosts.Host.\\d+.Active", "true")
       
        res.each do | o |
            prefix, match, subfix = o['path'].rpartition('.')
           
            physAddress = get_device_param_by_specific( @device_params, "#{prefix}.PhysAddress") 
            next if !physAddress.blank? && !physAddress['value'].blank? && @re_mac_list.include?(physAddress['value'])

            ipAddress = get_device_param_by_specific( @device_params, "#{prefix}.IPAddress")
            ipAddress = "0.0.0.0" if ipAddress.blank? && ipAddress['value'].blank?

            mac = get_device_param_by_specific( @device_params, "#{prefix}.X_TP_DevphyAddress")
           
            if !mac.blank? && !mac['value'].blank? && mac['value'] == mac_list
                lanConnType = get_device_param_by_specific( @device_params, "#{prefix}.X_TP_LanConnType") 
                if !lanConnType.blank?
                    case lanConnType['value']
                    when "0" #wired(0)
                        ret["wired"] += 1
                        type = "wired"
                    when "1" #wireless(1)
                        rssi = get_device_param_by_specific(@device_params, "#{prefix}.X_TP_Rssi")
                        next if rssi.blank? && !rssi['value'].blank?
                        level = get_rssi_level(rssi['value'].to_i)
                        ret[level] += 1
                    end
                end
            end
        end
        ret
    end

    def get_rssi_level(rssi)
        if rssi >= -50
          'good'
        elsif rssi >= -70 && rssi < -50
          'normal'
        elsif rssi < -70
          'poor'
        end
    end

    
    def getParameters()
        return @params.clone
    end
end
