#!/usr/bin/env node
/*  Copyright(c) 2020-2025 TP-LINK Technologies Co.Ltd.
 *
 * file      agileacs_activate_report.js
 * brief     nodejs command to update ativate report to las with http restful API
 * details   /usr/bin/nodejs /opt/genieacs/bin/agileacs_activate_report.js [action] [category] [item]
 *           [action]: collect || update
 *           [category]: activation , [item]: activation || deployment

 * author    Sharon Lin
 * version   1.0.0
 * date      21Apr20
 *
 * history   \arg 1.0.0, 21Apr20, Sharon Lin, Create the file.
 */

/**************************************************************************************************/
/*                                           DEFINES                                              */
/**************************************************************************************************/

/**************************************************************************************************/
/*                                           VARIABLES                                            */
/**************************************************************************************************/
var path = require('path');
var fs = require('fs');
var exec = require('child_process').exec;
var dir = path.resolve(path.dirname(fs.realpathSync(__filename)), '..');
process.chdir(dir);
const MongoClass = require('tplib/mongo-class')
const agileConfig = require('tplib/agile-config')
const stringTable = require('tplib/string-table.json')
const dbTplink = new MongoClass(agileConfig.get(stringTable.url.dbTplink))
const dbTmp = new MongoClass(agileConfig.get(stringTable.url.dbTmp))
const ACS_SYSTEM_LOG_FILE="/var/log/genieacs/acs-system.log"
const ACTIVATE_REPORT_DEBUG_FILE="/var/log/genieacs/activate_report.log"
const LICENSE_FEATURE_SET_ID="feature set license"
const LICENSE_DEFAULT_FEATURE_SET_ID="default feature set license"
const LICENSE_QUANTY_ID="quantity license"
const LICENSE_TRIAL_QUANTY_ID="trial"
const LICENSE_MAINTENANCE_ID="maintenance license"
const ACTIVATE_REPORT_DEFAULT_LASIP="https://Activation-AgileACS.TP-Link.com/api/v1/agileacs_update"
const ACTIVATE_REPORT_STATUS_NONE="none"
const ACTIVATE_REPORT_STATUS_ACTIVATING="activating"
const ACTIVATE_REPORT_STATUS_ACTIVATED="activated"
const ACTIVATE_REPORT_STATUS_FAILED="failed"
var args = process.argv.slice(2);
var action = args[0];
var category = args[1];
var item = args[2];
var enable_debug = true;

var las_api = require('./las-api');
/**************************************************************************************************/
/*                                           LOCAL_FUNCTIONS                                      */
/**************************************************************************************************/
Date.prototype.toIsoString = function() {
  function pad(number) {
    if (number < 10) {
      return '0' + number;
    }
    return number;
  }
  return this.getFullYear() +
      '-' + pad(this.getMonth() + 1) +
      '-' + pad(this.getDate()) +
      'T' + pad(this.getHours()) +
      ':' + pad(this.getMinutes()) +
      ':' + pad(this.getSeconds()) +
      '.' + (this.getMilliseconds() / 1000).toFixed(3).slice(2, 5) +
      'Z';
}

let log_level_str = function(level)
{
  var level_str="[DEBUG]";
  switch(level)
  {
    case 7:
        level_str="[DEBUG]";
        break;
    case 6:
        level_str="[INFO]";
        break;
    case 4:
        level_str="[WARN]";
        break;
    case 3:
        level_str="[ERROR]";
        break;
  }
  return level_str;
}

let log_time = function (level = 7, prefix)
{
  let level_str = log_level_str(level);
  let host = '127.0.0.1';
  let objectname = '[ActivateReport]:'
  let current_time = new Date().toIsoString();
  let log_str = current_time + ' ' + level_str+ ' ' + level + ' ' + host + ' ' + objectname + ' ' +  prefix;
  fs.appendFileSync(ACS_SYSTEM_LOG_FILE, log_str+"\n");
}

let log_debug = function (prefix)
{
  if(!enable_debug)
    return;
  let current_time = new Date().toIsoString();
  let log_str = current_time + ' ' +  prefix;
  fs.appendFileSync(ACTIVATE_REPORT_DEBUG_FILE, log_str+"\n");
  //console.log(log_str);
}

let http_send_data = function (las_url, data) {
  //to do send data
  return new Promise(function(resolve){
    las_api.sendToLAS(las_url, data, 0)
    .then((res) => {
      return resolve(res);
    })
    .catch((err) => {
      return resolve(err);
    })
  })
}

let update_activate_report_status = function(status) {
  return new Promise(function(resolve){
    cmd='/opt/agileacs/bin/acsutils2 activatereport update status '+ status;
    log_debug("[update_activate_report_status] cmd:" + cmd);
    return exec(cmd ,function( err, stdout, stderr){
      if(err) {
        log_debug("[update_activate_report_status]  err:" + err);
      }
      if(stderr) {
        log_debug("[update_activate_report_status]  stderr:" +stderr);
      }
      log_debug("[update_activate_report_status]  stdout:" +stdout);
      return resolve();
    });
  })
}

let get_outage_threshold = function (dbTplink) {
  return new Promise((resolve) => {
    dbTplink.findOne(stringTable.collection.reportSetting, {_id: "report_setting"}, function(err,result){
      log_debug("report_setting outage_threshold = " + result.outage_threshold);
      resolve(result.outage_threshold);
    })
  })
}

let get_serverinfo = function () {
  return new Promise((resolve, reject) => {
    let serverinfo_collect = () => {
      exec('/opt/agileacs/bin/acsutils2 collect 2>&1' ,function(err,stdout,stderr){
        var contents, jsonContenct;
        if(err) {
            log_debug("[get_serverinfo] serverinfo_collect err:" + err);
            setTimeout(() => serverinfo_collect(), 1000)
        }
        if(stderr) {
            log_debug("[get_serverinfo] serverinfo_collect stderr:" +stderr);
            setTimeout(() => serverinfo_collect(), 1000)
        }
        if(stdout) {
          contents = fs.readFileSync(stdout);
          jsonContenct = JSON.parse(contents);
        }
        log_debug("[get_serverinfo] jsonContenct =" + JSON.stringify(jsonContenct));
        resolve(jsonContenct);
      })
    }
    serverinfo_collect();
  })
}

let get_activation_report_setting = function(dbTplink) {
  return new Promise((resolve) => {
    var proj = {
      'ceip_enabled':1,
      'ceip_interval':1,
      'las_url':1,
      'status':1,
    };
    return dbTplink.findOne(stringTable.collection.activateReport, {_id: "service"}, proj, function(err,result){
      var setting = {
        'ceip_enabled':1,
        'ceip_interval':7,
        'las_url': ACTIVATE_REPORT_DEFAULT_LASIP,
        'status': ACTIVATE_REPORT_STATUS_NONE
      };
      if(err)
      {
          log_debug("[get_activation_report_setting] collection findOne err: "+err);
          log_time(4, "[get_activation_report_setting]  collection findOne err: "+err);
          return resolve();
      }
      if(result.hasOwnProperty('ceip_enabled'))
      setting['ceip_enabled'] = result['ceip_enabled'];

      if(result.hasOwnProperty('ceip_interval'))
      setting['ceip_interval'] = result['ceip_interval'];

      if(result.hasOwnProperty('las_url'))
      setting['las_url'] = result['las_url'];

      if(result.hasOwnProperty('status'))
      setting['status'] = result['status'];

      log_debug("[get_activation_report_setting] setting = " + JSON.stringify(setting));
      return resolve(setting);
    });
  })
}

let get_wispinfo_setting = function(dbTplink) {
  return new Promise((resolve) => {
    var proj = {
      'wisp_name':1,
      'wisp_email':1,
      'wisp_country':1,
      'wisp_region':1,
      'wisp_cnpj':1
    };
    return dbTplink.findOne(stringTable.collection.activateReport, {_id: "service"}, proj, function(err,result){
      var wispinfo = {
        'name':"",
        'email':"",
        'country':"",
        'region':"",
        'cnpj':""
      };
      if(err)
      {
          log_debug("[get_wispinfo_setting] collection findOne err: "+err);
          log_time(4, "[get_wispinfo_setting]  collection findOne err: "+err);
          return resolve();
      }
      if(result.hasOwnProperty('wisp_name'))
      wispinfo['name'] = result['wisp_name'];

      if(result.hasOwnProperty('wisp_email'))
      wispinfo['email'] = result['wisp_email'];

      if(result.hasOwnProperty('wisp_country'))
      wispinfo['country'] = result['wisp_country'];

      if(result.hasOwnProperty('wisp_region'))
      wispinfo['region'] = result['wisp_region'];

      if(result.hasOwnProperty('wisp_cnpj'))
      wispinfo['cnpj'] = result['wisp_cnpj'];
      log_debug("[get_wispinfo_setting] wispinfo = " + JSON.stringify(wispinfo));
      return resolve(wispinfo);
    });
  })
}

let get_activation_license = function(dbTplink) {
  return new Promise((resolve) => {
    return dbTplink.find(stringTable.collection.totalLicense, {}, function(err,results){
      data = {
          'feature_set':[],
          'quantity':[],
          'maintenance':[],
        };
        var i;
        var feature_set_data, quantity_data, maintenance_data;
        if(err)
        {
            log_debug("[get_activation_license] collection find err: "+err);
            log_time(4,"[get_activation_license] collection find err: "+err);
            return resolve();
        }

        if(results && results.length)
        {
          for(i=0; i<results.length; i++)
          {
            if(results[i]['_id'] === LICENSE_FEATURE_SET_ID || results[i]['_id'] === LICENSE_DEFAULT_FEATURE_SET_ID) //feature set
            {
              feature_set_data={};
              feature_set_data['last_activated_date'] = results[i]['activated_date'];
              feature_set_data['feature_set'] = results[i]['feature_set'];
              if(results[i]['expireddays'] === -1)
              {
                feature_set_data['expired_in_days'] = -1;
                feature_set_data['expired_date'] = -1;
              }
              else
              {
                feature_set_data['expired_in_days'] = results[i]['expireddays'];
                feature_set_data['expired_date'] = new Date(results[i]['expired_timestamp']*1000);
              }
              feature_set_data['status'] = results[i]['status'];
              data['feature_set'].push(feature_set_data);
            }
            if(results[i]['_id'] === LICENSE_QUANTY_ID || results[i]['_id'] === LICENSE_TRIAL_QUANTY_ID) //quantity
            {
              quantity_data={};
              quantity_data['last_activated_date'] = results[i]['activated_date'];
              quantity_data['controlnum'] = results[i]['controlnum'];
              quantity_data['used'] = results[i]['controlnum'] - results[i]['available_controlnum'];
              quantity_data['available'] = results[i]['available_controlnum'];
              if(results[i]['expireddays'] === -1)
              {
                quantity_data['expired_in_days'] = -1;
                quantity_data['expired_date'] = -1;
              }
              else
              {
                quantity_data['expired_in_days'] = results[i]['expireddays'];
                quantity_data['expired_date'] = new Date(results[i]['expired_timestamp']*1000);
              }
              quantity_data['status'] = results[i]['status'];
              data['quantity'].push(quantity_data);
            }
            if(results[i]['_id'] === LICENSE_MAINTENANCE_ID ) //maintenance
            {
              maintenance_data={};
              maintenance_data['last_activated_date'] = results[i]['activated_date'];
              if(results[i]['expireddays'] === -1)
              {
                maintenance_data['expired_in_days'] = -1;
                maintenance_data['expired_date'] = -1;
              }
              else
              {
                maintenance_data['expired_in_days'] = results[i]['expireddays'];
                maintenance_data['expired_date'] = new Date(results[i]['expired_timestamp']*1000);
              }
              data['maintenance'].push(maintenance_data);
            }
          }
        }
        log_debug("[get_activation_license] data = " + JSON.stringify(data));
        return resolve(data);
    });
  })
}

let get_activation_inventory = function(dbTmp, dbTplink) {

  return new Promise(async (resolve) => {
    let bound = new Date();
    let threshold = await get_outage_threshold(dbTplink)
    bound.setDate(bound.getDate()-threshold);
    return dbTmp.aggregate(stringTable.collection.devices, [{"$match":{"_lastInform":{"$gte": bound}}},{$project: {ModelName: { $ifNull :[ "$Device.DeviceInfo.ModelName._value","$InternetGatewayDevice.DeviceInfo.ModelName._value"]}}}]).toArray(function(err,devices){
      var i =0, j =0, modelName;
      var inventory=[];
      if(err)
      {
          log_debug("[get_activation_inventory] collection $gte aggregate err: "+err);
          log_time(4,"[get_activation_inventory] collection $gte aggregate err: "+err);
          return resolve();
      }
      if(devices && devices.length)
      {
        i =0; j =0;
        while(devices[i])
        {
          modelName = devices[i].ModelName;
          for(j=0; inventory[j];j++) {
            if(inventory[j]["model"] === modelName) {
              inventory[j]["registered"] ++;
              inventory[j]["online"] ++;
              break;
            }
          }
          if(!inventory[j]){//Means no model name satisfied modelName in report date
            inventory[j] = {'model': modelName,'registered': 1, 'online':1, 'outage':0};
          }
          i++;
        }
      }
      return dbTmp.aggregate(stringTable.collection.devices,[{"$match":{"_lastInform":{"$lt": bound}}},{$project: {ModelName: { $ifNull :[ "$Device.DeviceInfo.ModelName._value","$InternetGatewayDevice.DeviceInfo.ModelName._value"]}}}]).toArray(function(err,devices){
        if(err)
        {
          log_debug("[get_activation_inventory] collection $lt aggregate err: "+err);
          log_time(4,"[get_activation_inventory] collection $lt aggregate err: "+err);
          return resolve();
        }
        if(devices && devices.length)
        {
          i =0; j =0;
          while(devices[i])
          {
            modelName = devices[i].ModelName;
            for(j=0; inventory[j];j++) {
              if(inventory[j]["model"] === modelName) {
                inventory[j]["registered"] ++;
                inventory[j]["outage"] ++;
                break;
              }
            }
            if(!inventory[j]){//Means no model name satisfied modelName in report date
              inventory[j] = {'model': modelName,'registered': 1, 'online':0, 'outage':1};
            }
            i++;
          }
        }
        log_debug("[get_activation_inventory] inventory = " + JSON.stringify(inventory));
        return resolve(inventory);
      })
    })
  })
}

let get_activation_ativation_data = function(dbTmp, dbTplink, category, item) {
  return Promise.all([
    get_serverinfo(),
    get_wispinfo_setting(dbTplink)
  ])
  .then((results) => {
    var data={
      'category': category,
      'item': item,
      'serverinfo':{},
      'wispinfo':{}
    };
    if(Object.keys(results[0]).length !== 0)
      data['serverinfo'] = results[0];
    if(Object.keys(results[1]).length !== 0)
      data['wispinfo'] = results[1];
    //log_debug("data: " + JSON.stringify(data));
    return data;
  })
}

let get_activation_deployment_data = function(dbTmp, dbTplink, category, item) {
  return Promise.all([
    get_serverinfo(),
    get_wispinfo_setting(dbTplink),
    get_activation_license(dbTplink),
    get_activation_inventory(dbTmp, dbTplink)
  ])
  .then((results) => {
    var data={
      'category': category,
      'item': item,
      'serverinfo':{},
      'wispinfo':{},
      'license':{},
      'inventory':{}
    };
    if(Object.keys(results[0]).length !== 0)
      data['serverinfo'] = results[0];
    if(Object.keys(results[1]).length !== 0)
      data['wispinfo'] = results[1];
    if(Object.keys(results[2]).length !== 0)
      data['license'] = results[2];
    if(results[3])
      data['inventory'] = results[3];
    //log_debug("data: " + JSON.stringify(data));
    return data;
  })
}

let get_update_data = function(dbTmp, dbTplink, category, item) {
  switch(category)
  {
    case 'activation':
      if(item === 'activation')
        return get_activation_ativation_data(dbTmp, dbTplink, category, item);
      else if(item === 'deployment')
        return get_activation_deployment_data(dbTmp, dbTplink, category, item);
    break;
  }
  return ;
}

let send_update = function(category, item, setting, data) {
  return new Promise(async function(resolve){
    if(category === 'activation' && item === 'deployment' && 
    ( setting['ceip_enabled'] === 0 || setting['status'] !== ACTIVATE_REPORT_STATUS_ACTIVATED)){
      if(setting['ceip_enabled'] === 0)
        log_time(7, "[" + category + ":" + item + "] disabled");
      if(setting['status'] !== ACTIVATE_REPORT_STATUS_ACTIVATED)
        log_time(7, "[" + category + ":" + item + "] activation not ready yet");
      return resolve(0);
    }
    if(category === 'activation' && item === 'activation' && setting['status'] === ACTIVATE_REPORT_STATUS_ACTIVATED){
      log_debug("[" + category + ":" + item + "] has been activated already");
      return resolve(2);
    }
    log_debug("send [" + category + ":" + item + "] data to server [" +setting['las_url']+"]");
    if(category === 'activation' && item === 'activation')
      await update_activate_report_status(ACTIVATE_REPORT_STATUS_ACTIVATING);
    let ret = await http_send_data(setting['las_url'], data);
    if(ret.hasOwnProperty('status') && ret['status'] === '200')
    {
      log_time(6, "send [" + category + ":" + item + "] data succeed.");
      if(category === 'activation' && item === 'activation')
       await update_activate_report_status(ACTIVATE_REPORT_STATUS_ACTIVATED);
       return resolve(1);
    }
    else
    {
      if(ret.hasOwnProperty('message'))
        log_time(4, "send failed " + ret['message']);
      log_time(4, "send [" + category + ":" + item + "] data failed.");
      if(category === 'activation' && item === 'activation')
        await update_activate_report_status(ACTIVATE_REPORT_STATUS_FAILED);
        return resolve(0);
    }
  })
}

let update = function(category, item) {
  var setting, data;
  Promise.all([
    dbTplink.waitConnected(),
    dbTmp.waitConnected()
  ])
  .then(() =>{
      Promise.all([
        get_activation_report_setting(dbTplink),
        get_update_data(dbTmp, dbTplink, category, item)
      ])
      .then((results)=>{

          if(results[0])
            setting = results[0];
          if(results[1])
            data = results[1];
          send_update(category, item, setting, data)
          .then((ret)=>{
            console.log(ret);
            log_debug("[update] return ret = " + ret);
            process.exit(0);
          })
      })
      .catch((err) => {
        log_debug("[update] err: " + err);
        console.log("0");
        log_debug("[update] return ret = 0");
        process.exit(0);
      })
    })
  .catch(() =>{
      log_debug("Cannot connect to dbTplink & dbTmp");
      console.log("0");
      log_debug("[update] Cannot connect to dbTplink & dbTmp return ret = 0");
      process.exit(0);
  })
}

let test = async function(category, item) {
  log_time(6, "send [" + category + ":" + item + "] data succeed.");
  await update_activate_report_status(ACTIVATE_REPORT_STATUS_ACTIVATED);
  console.log("1");
  process.exit(0);
}

if (action==='update') {
  if (category==='activation') {
    if(item==='activation'|| item==='deployment') {
      return update(category, item);
    }
  }
} else if (action==='collection') {
  //to do
} else if (action==='test') {
  return test('test','test');
}

log_debug("Unknow args " + args);
console.log("0");
process.exit(0);
