#!/usr/bin/env node
/*  Copyright(c) 2009-2019 TP-LINK Technologies Co.Ltd.
 *
 * file      agileacs_report.js
 * brief     nodejs command to generate monthly report
 * details   /usr/bin/nodejs /opt/genieacs/bin/agileacs_report.js [action] [item]
 *           [action]: collect || update
 *                     collect => [item]: all || reportsetting
 *                                update => [item]: cpeinventory || useroperation || profilestats || invaliddevice
 * author    Hsinhao Yao
 * version   1.0.0
 * date      31July19
 *
 * history   \arg 1.0.0, 31July19, Hsinhao Yao, Create the file.
 */

/**************************************************************************************************/
/*                                           DEFINES                                              */
/**************************************************************************************************/

var report_config = { "storage_period" : 6};

/**************************************************************************************************/
/*                                           VARIABLES                                            */
/**************************************************************************************************/
var path = require('path');
var fs = require('fs');
var dir = path.resolve(path.dirname(fs.realpathSync(__filename)), '..');
process.chdir(dir);
const MongoClass = require('tplib/mongo-class')
const agileConfig = require('tplib/agile-config')
const stringTable = require('tplib/string-table.json')
const dbTplink = new MongoClass(agileConfig.get(stringTable.url.dbTplink))
const dbGenie = new MongoClass(agileConfig.get(stringTable.url.dbGenie))
const dbTmp = new MongoClass(agileConfig.get(stringTable.url.dbTmp))
const MONTH_SECONDS = 2678400000;

var args = process.argv.slice(2);
var action = args[0];
var items = args[1];

var now = new Date();
let month = now.getMonth()+1;
let year = now.getFullYear();
let day = now.getDate();
if(day<10){
  day = "0"+day;
}
let deleteTime = new Date();
deleteTime.setMonth(deleteTime.getMonth()-report_config.storage_period);
let deleteMonth = deleteTime.getMonth()+1;

/**************************************************************************************************/
/*                                           LOCAL_FUNCTIONS                                      */
/**************************************************************************************************/
/*
 * fn        function IsProFeatureSet()
 * brief     
 * details   
 *
 * param[in] callback(boolean)
 *
 * return    N/A
 * retval
 * console   
 *
 * note      Written by Hsinhao. 2019-10-14
 */
function IsProFeatureSet(dbTplink, callback) {
  dbTplink.findOne(stringTable.collection.totalLicense, {_id: "feature set license"}, function(err, result){//Get report_setting.outage_threshold
    if(err) return callback(false);
    if(result === null) return callback(false);
    if(result.feature_set === "pro-trial"){
      if(result.status === "activated")
        return callback(true);
    } else if(result.feature_set === "pro"){
      if(result.status === "activated")
        return callback(true);
    }
    return callback(false);
  });
}

/*
 * fn        function updateCpeInventory()
 * brief     update cpe inventory report
 * details   get _lastInform and ModelName from devices collection
 *           get outage_threshold from report_setting collection
 *           _lastInform is less than outage_threshold days is online
 *           _lastInform is greater than outage_threshold days is outage
 *           generate report like: [{'model': modelName,'registered': n+m, 'online' : n, 'outage' : m},...]
 * param[in] N/A
 *
 * return    N/A
 * retval
 * console   reportPeriodStart.valueOf() + "," + now.valueOf()
 *
 * note      Written by Hsinhao. 2019-7-31
 */
function updateCpeInventory(dbGenie, dbTplink, dbTmp) {
  var cpeInventory = [];
  let bound = new Date();
  let localMonth = month;
  let localYear = year;
  let localDeleteMonth = deleteMonth;

  return new Promise( (resolve, reject) => {
    dbTplink.findOne(stringTable.collection.reportSetting, {_id: "report_setting"}, function(err, result){//Get report_setting.outage_threshold
      if(err) reject("[updateCpeInventory] dbTplink.findOne reportSetting error")
      bound.setDate(bound.getDate()-result.outage_threshold);
      if(result.reportday === "end"){
      let thisMonthEndDate = new Date(now.getYear()+1900, now.getMonth()+1, 1, 0, 0, 0);
        thisMonthEndDate.setDate(thisMonthEndDate.getDate()-1);//Get end date of this month
        if(now.getDate()>=thisMonthEndDate.getDate()) {//next report period
          localMonth ++;
          localDeleteMonth ++;
        }
      } else {
        if(now.getDate()>=parseInt(result.reportday, 10)){
          localMonth++;
          localDeleteMonth++;
        }
      }
      if(localMonth>12){
        localMonth-=12;
        localYear++;
      }
      if(localMonth<10){
        localMonth = "0"+localMonth;
      }
      if(localDeleteMonth<10){
        localDeleteMonth = "0"+localDeleteMonth;
      }

      let id = "" + localYear + localMonth;
      let deleteId = "" + deleteTime.getFullYear() + localDeleteMonth;
      dbTmp.aggregate(stringTable.collection.devices, [{"$match":{"_lastInform":{"$gte": bound}}},{$project: {ModelName: { $ifNull :[ "$Device.DeviceInfo.ModelName._value","$InternetGatewayDevice.DeviceInfo.ModelName._value"]}}}]).toArray(function(err,result){
        //get devices.ModelName while devices._lastInform >= (now-7days)
        //console.log(result);
        if(err) reject("[updateCpeInventory] dbTmp.findOne aggregate devices error")
        var i =0, j =0;
        while(result[i])
        {
          modelName = result[i].ModelName;
          for(j=0; cpeInventory[j];j++) {
            if(cpeInventory[j]["model"] === modelName) {
              cpeInventory[j]["online"] ++;
              cpeInventory[j]["registered"] ++;
              break;
            }
          }
          if(!cpeInventory[j]){//Means no model name satisfied modelName in report date
            cpeInventory[j] = {'model': modelName,'registered': 1, 'online' : 1, 'outage' : 0};
          }
          i++;
        }
        dbTmp.aggregate(stringTable.collection.devices, [{"$match":{"_lastInform":{"$lt": bound}}},{$project: {ModelName: { $ifNull :[ "$Device.DeviceInfo.ModelName._value","$InternetGatewayDevice.DeviceInfo.ModelName._value"]}}}]).toArray(function(err,result){
          //get devices.ModelName while devices._lastInform < (now-7days)
          if(err) reject("[updateCpeInventory] dbTmp.findOne aggregate devices error")
          var i =0, j =0;
          while(result[i])
          {
            modelName = result[i].ModelName;
            for(j=0; cpeInventory[j];j++) {
              if(cpeInventory[j]["model"] === modelName) {
                cpeInventory[j]["outage"] ++;
                cpeInventory[j]["registered"] ++;
                break;
              }
            }
            if(!cpeInventory[j]){//Means no model name satisfied modelName in report date
              cpeInventory[j] = {'model': modelName,'registered': 1, 'online' : 0, 'outage' : 1};
            }
            i++;
          }
          //console.log("[updateCpeInventory] cpeInventory = " + JSON.stringify(cpeInventory));
          IsProFeatureSet(dbTplink, function(isPro){
            if(isPro){
              dbTplink.deleteOne(stringTable.collection.cpeInventoryReport, {_id: id},function(){//Remove current data if exists
                dbTplink.insertOne(stringTable.collection.cpeInventoryReport, {_id: id, "report": cpeInventory, "timestamp": now.valueOf()}, function (err, res){
                  if(err) reject("[updateCpeInventory] dbTplink.insertOne " + collname + " error")
                  //console.log("[updateCpeInventory] res = " + JSON.stringify(res));
                  dbTplink.deleteOne(stringTable.collection.cpeInventoryReport, {_id: deleteId},function(){//Remove current data if exists
                    if(err) reject("[updateCpeInventory] dbTplink.deleteOne " + stringTable.collection.cpeInventoryReport + " error")
                    collectReportSetting(dbGenie, dbTplink, dbTmp)
                      .then((res) => {
                        resolve(res)
                    })
                    .catch((err) => {
                      reject("[collectReportSetting] Error: " + err)
                    })
                  });
                });
              });
            }
            else
              reject("[updateCpeInventory] Feature Set is not Pro")
          });
        });
      });
    });
  });
}

/*
 * fn        function updateUserOperation()
 * brief     update user operation report
 * details   get data from tmpuseroperation collection
 *           get reportday from report_setting collection
 *           determine the report period by using reportday
 *           generate report like: [{"operation":"login","user":"admin","succeed":m,"failed":n},{"operation":"change_password","user":"admin","succeed":k,"failed":l}]
 * param[in] N/A
 *
 * return    N/A
 * retval
 * console   reportPeriodStart.valueOf() + "," + now.valueOf()
 *
 * note      Written by Hsinhao. 2019-7-31
 */
function updateUserOperation(dbGenie, dbTplink, dbTmp) {
    var userOperation = [];
    let bound, id, deleteId;
    let localMonth = month;
    let localYear = year;
    let localDeleteMonth = deleteMonth;

    return new Promise( (resolve, reject) => {
      dbTplink.findOne(stringTable.collection.reportSetting, {_id: "report_setting"}, function(err, result){//Get report_setting.outage_threshold
        if(err) reject("[updateUserOperation] dbTplink.findOne reportSetting error")
        //console.log(result.reportday);
        if(result.reportday === "end"){
          let thisMonthEndDate = new Date(now.getYear()+1900, now.getMonth()+1, 1, 0, 0, 0);
          thisMonthEndDate.setDate(thisMonthEndDate.getDate()-1);//Get end date of this month
          if(now.getDate()>=thisMonthEndDate.getDate()) {//next report period
            bound = thisMonthEndDate;//Remove period: before end date of this month
            localMonth ++;
            localDeleteMonth ++;
          } else {
            bound = new Date(now.getYear()+1900, now.getMonth(), 1, 0, 0, 0);
            bound.setDate(bound.getDate()-1);//Remove period: before end date of last month
          }
        } else {
          if(now.getDate()>=parseInt(result.reportday, 10)){//next report period
            bound = new Date(now.getYear()+1900, now.getMonth(), parseInt(result.reportday, 10), 0, 0, 0); //Remove period: before $reportday date of this month
            localMonth ++;
            localDeleteMonth ++;
          } else {
            bound = new Date(now.getYear()+1900, now.getMonth()-1, parseInt(result.reportday, 10), 0, 0, 0); //Remove period: before $reportday date of last month
          }
        }
        if(localMonth>12){
          localMonth-=12;
          localYear++;
        }
        if(localDeleteMonth<10){
          localDeleteMonth = "0"+localDeleteMonth;
        }
        if(localMonth<10){
          localMonth = "0"+localMonth;
        }
        id = "" + localYear + localMonth;
        deleteId = "" + deleteTime.getFullYear() + localDeleteMonth;
        // console.log(bound.valueOf()+ "," +now.valueOf());

        dbTplink.deleteMany(stringTable.collection.tmpUserOperation,{"timestamp": {"$lt": bound.valueOf() - MONTH_SECONDS}},function(){
          dbTplink.find(stringTable.collection.tmpUserOperation,{"timestamp": {"$gte": bound.valueOf()}},function(err,result){
            if(err) reject("[updateUserOperation] dbTplink.find tmpUserOperation error")
            for (let obj of result)
            {
              let index = userOperation.findIndex(v => v.user == obj.user && v.operation == obj.operation);
              if (index == -1) {
                userOperation.push({ "operation": obj.operation, "user": obj.user, "succeed": 0, "failed": 0 })
                index = userOperation.findIndex(v => v.user == obj.user && v.operation == obj.operation);
              }
              if (obj.status === 1) {
                userOperation[index].succeed ++;
              } else {
                userOperation[index].failed ++;
              }
            }
            //console.log("[updateUserOperation] userOperation = " + JSON.stringify(userOperation));
            IsProFeatureSet(dbTplink, function(isPro){
              if(isPro){
                dbTplink.deleteOne(stringTable.collection.userOperationReport, {_id: id},function(){//Remove current data if exists
                  dbTplink.insertOne(stringTable.collection.userOperationReport, {_id: id, "report": userOperation, "timestamp": now.valueOf()}, function (err, res){
                    if(err) reject("[updateUserOperation] dbTplink.insertOne " + collname + " error")
                    //console.log("[updateUserOperation] res = " + JSON.stringify(res));
                    dbTplink.deleteOne(stringTable.collection.userOperationReport, {_id: deleteId},function(){//Remove current data if exists
                      if(err) reject("[updateUserOperation] dbTplink.deleteOne " + stringTable.collection.userOperationReport + " error")
                      resolve(bound.valueOf() + "," + now.valueOf());
                    });
                  });
                });
              }
              else
                reject("[updateUserOperation] Feature Set is not Pro");
            });
          });
        });
      });
    });
}

/*
 * fn        function updateProfileStats()
 * brief     update profile statistic report
 * details   get data from tmpprofilestats collection
 *           get reportday from report_setting collection
 *           determine the report period by using reportday
 *           generate report like: [{'profile': profileName, 'model': modelName, 'app': appName, 'complete': n, 'error' : m},...]
 *           appName contains: Firmware Upgrade, Configure File Download, PPPoE, and Periodic Inform Interval
 * param[in] N/A
 *
 * return    N/A
 * retval
 * console   reportPeriodStart.valueOf() + "," + now.valueOf()
 *
 * note      Written by Hsinhao. 2019-7-31
 */
function updateProfileStats(dbGenie, dbTplink, dbTmp) {
    var profileStats = [];
    let bound, id, deleteId;
    let localMonth = month;
    let localYear = year;
    let localDeleteMonth = deleteMonth;

    return new Promise( (resolve, reject) => {
      dbTplink.findOne(stringTable.collection.reportSetting, {_id: "report_setting"}, function(err, result){//Get report_setting.outage_threshold
        if(err) reject("[updateProfileStats] dbTplink.findOne reportSetting error")
        //console.log(result.reportday);
        if(result.reportday === "end"){
          let thisMonthEndDate = new Date(now.getYear()+1900, now.getMonth()+1, 1, 0, 0, 0);
          thisMonthEndDate.setDate(thisMonthEndDate.getDate()-1);//Get end date of this month
          if(now.getDate()>=thisMonthEndDate.getDate()) {//next report period
            bound = thisMonthEndDate;//Remove period: before end date of this month
            localMonth ++;
            localDeleteMonth ++;
          } else {
            bound = new Date(now.getYear()+1900, now.getMonth(), 1, 0, 0, 0);
            bound.setDate(bound.getDate()-1);//Remove period: before end date of last month
          }
        } else {
          if(now.getDate()>=parseInt(result.reportday, 10)){//next report period
            bound = new Date(now.getYear()+1900, now.getMonth(), parseInt(result.reportday, 10), 0, 0, 0); //Remove period: before $reportday date of this month
            localMonth ++;
            localDeleteMonth ++;
          } else {
            bound = new Date(now.getYear()+1900, now.getMonth()-1, parseInt(result.reportday, 10), 0, 0, 0); //Remove period: before $reportday date of last month
          }
        }
        if(localMonth>12){
          localMonth-=12;
          localYear++;
        }
        if(localDeleteMonth<10){
          localDeleteMonth = "0"+localDeleteMonth;
        }
        if(localMonth<10){
          localMonth = "0"+localMonth;
        }
        id = "" + localYear + localMonth;
        deleteId = "" + deleteTime.getFullYear() + localDeleteMonth;
        //console.log(bound.valueOf()+ "," +now.valueOf());
        dbTplink.deleteMany(stringTable.collection.tmpProfileStats,{"timestamp": {"$lt": bound.valueOf() - MONTH_SECONDS}},function(){
          dbTplink.find(stringTable.collection.tmpProfileStats,{"timestamp": {"$gte": bound.valueOf()}},function(err,result){
            if(err) reject("[updateProfileStats] dbTplink.find tmpProfileStats error")
            var i =0, j =0, profileName;
            while(result[i])
            {
              profileName = result[i].name;
              for(j=0; profileStats[j];j++) {
                if(profileStats[j]["profile"] === profileName) {
                  if(result[i]["status"] === "2"){
                    profileStats[j]["complete"] ++;
                  } else {
                    profileStats[j]["error"] ++;
                  }
                  break;
                }
              }
              if(!profileStats[j]){//Means no profile name satisfied profileName in report date
                if(result[i]["status"] === "2"){
                  profileStats[j] = {'profile': profileName, 'model': result[i]["model"], 'app': result[i]["app"], 'complete': 1, 'error' : 0};
                } else {
                  profileStats[j] = {'profile': profileName, 'model': result[i]["model"], 'app': result[i]["app"], 'complete': 0, 'error' : 1};
                }
              }
              i++;
            }
            //console.log("[updateProfileStats] profileStats = " + JSON.stringify(profileStats));
            IsProFeatureSet(dbTplink, function(isPro){
              if(isPro){
                dbTplink.deleteOne(stringTable.collection.profileStatsReport, {_id: id},function(){//Remove current data if exists
                  dbTplink.insertOne(stringTable.collection.profileStatsReport, {_id: id, "report": profileStats, "timestamp": now.valueOf()}, function (err, res){
                    if(err) reject("[updateProfileStats] dbTplink.insertOne " + collname + " error")
                   // console.log("[updateProfileStats] res = " + JSON.stringify(res));
                    dbTplink.deleteOne(stringTable.collection.profileStatsReport, {_id: deleteId},function(){//Remove current data if exists
                      if(err) reject("[updateProfileStats] dbTplink.deleteOne " + stringTable.collection.profileStatsReport + " error")
                      resolve(bound.valueOf() + "," + now.valueOf());
                    });
                  });
                });
              }
              else
                reject("[updateProfileStats] Feature Set is not Pro");
            });
          });
        });
      });
    });
}

/*
 * fn        function updateInvalidDevice()
 * brief     update invalid device report
 * details   get data from tmpinvalidevice collection
 *           get reportday from report_setting collection
 *           determine the report period by using reportday
 *           generate report like: [{'model': modelName,'expired': n, 'invalid' : m},...]
 *           store to mongodb everyday
 * param[in] N/A
 *
 * return    N/A
 * retval
 * console   reportPeriodStart.valueOf() + "," + now.valueOf()
 *
 * note      Written by Hsinhao. 2019-7-31
 */
function updateInvalidDevice(dbGenie, dbTplink, dbTmp) {
  var invalidDevice = [];
  let localMonth = month;
  let localYear = year;
  if(localMonth>12){
    localMonth-=12;
    localYear++;
  }
  if(localMonth<10){
    localMonth = "0"+localMonth;
  }
  let id = "" + localYear + localMonth + day;
  let reportPeriod;
  let bound = new Date(now.getYear()+1900, now.getMonth(), now.getDate(), 0, 0, 0);

  return new Promise( (resolve, reject) => {
    dbTplink.findOne(stringTable.collection.reportSetting, {_id: "report_setting"}, function(err, result){//Get report_setting.outage_threshold
      if(err) reject("[updateInvalidDevice] dbTplink.findOne reportSetting error")
      if(result.reportday === "end"){
        let thisMonthEndDate = new Date(now.getYear()+1900, now.getMonth()+1, 1, 0, 0, 0);
        thisMonthEndDate.setDate(thisMonthEndDate.getDate()-1);//Get end date of this month
        if(now.getDate()>=thisMonthEndDate.getDate()) {//next report period
          reportPeriod = thisMonthEndDate;//Remove period: before end date of this month
        } else {
          reportPeriod = new Date(now.getYear()+1900, now.getMonth(), 1, 0, 0, 0);
          reportPeriod.setDate(reportPeriod.getDate()-1);//Remove period: before end date of last month
        }
      } else {
        if(now.getDate()>=parseInt(result.reportday, 10)){//next report period
          reportPeriod = new Date(now.getYear()+1900, now.getMonth(), parseInt(result.reportday, 10), 0, 0, 0);//Remove period: before $reportday date of this month
        } else {
          reportPeriod = new Date(now.getYear()+1900, now.getMonth()-1, parseInt(result.reportday, 10), 0, 0, 0);//Remove period: before $reportday date of last month
        }
      }
      //console.log(reportPeriod.valueOf() + "," + now.valueOf());
      dbGenie.deleteMany(stringTable.collection.tmpInvalidDevice,{"timestamp": {"$lt": bound.valueOf() - MONTH_SECONDS}},function(){
        dbGenie.find(stringTable.collection.tmpInvalidDevice,{"timestamp": {"$gte": bound.valueOf()}},function(err,result){
          if(err) reject("[updateInvalidDevice] dbTplink.find tmpInvalidDevice error")
          var i =0, j =0, modelName;
          while(result[i])
          {
            modelName = result[i].model;
            for(j=0; invalidDevice[j];j++) {
              if(invalidDevice[j]["model"] === modelName) {
                if(result[i]["status"] === 2){
                  invalidDevice[j]["expired"] ++;
                } else {
                  invalidDevice[j]["invalid"] ++;
                }
                break;
              }
            }
            if(!invalidDevice[j]){//Means no model name satisfied modelName in report date
              if(result[i]["status"] === 2){
                invalidDevice[j] = {'model': modelName,'expired': 1, 'invalid' : 0};
              } else {
                invalidDevice[j] = {'model': modelName,'expired': 0, 'invalid' : 1};
              }
            }
            i++;
          }
          //console.log("[updateInvalidDevice] invalidDevice = " + JSON.stringify(invalidDevice));
          IsProFeatureSet(dbTplink, function(isPro){
            if(isPro){
              dbTplink.deleteOne(stringTable.collection.invalidDeviceReport, {_id: id},function(){//Remove current data if exists
                dbTplink.insertOne(stringTable.collection.invalidDeviceReport,{_id: id,"report":invalidDevice,"timestamp": now.valueOf()}, function (err, res){
                  if(err) reject("[updateInvalidDevice] dbTplink.insertOne invalidDeviceReport error")
                  //console.log("[updateInvalidDevice] res + " + JSON.stringify(res));
                  let deleteReportPeriod = new Date(reportPeriod.valueOf());
                  deleteReportPeriod.setMonth(deleteReportPeriod.getMonth()-report_config.storage_period+1);
                  dbTplink.deleteMany(stringTable.collection.invalidDeviceReport, { "timestamp": { "$lt": deleteReportPeriod.valueOf()}}, function(err){//Remove out of storage_period if exists
                    if(err) reject("[updateInvalidDevice] dbTplink.deleteMany invalidDeviceReport error")
                    resolve(reportPeriod.valueOf() + "," + now.valueOf());
                  });
                });
              });
            }
            else
              reject("[updateInvalidDevice] Feature Set is not Pro");
          });
        });
      });
    });
  });
}

/*
 * fn        function collectReportSetting()
 * brief     get report period
 * details
 * param[in] N/A
 *
 * return    N/A
 * retval
 * console   reportPeriodStart.valueOf() + "," + now.valueOf()
 *
 * note      Written by Hsinhao. 2019-7-31
 */
function collectReportSetting(dbGenie, dbTplink, dbTmp) {
  let reportPeriod;
  //console.log("[collectReportSetting] Start ");
  return new Promise( (resolve, reject) => {
    dbTplink.findOne(stringTable.collection.reportSetting, {_id: "report_setting"}, function(err, result){//Get report_setting.outage_threshold
      if(err) reject("[collectReportSetting ] dbTplink.findOne reportSetting error");
      if(result.reportday === "end"){
        let thisMonthEndDate = new Date(now.getYear()+1900, now.getMonth()+1, 1, 0, 0, 0);
        thisMonthEndDate.setDate(thisMonthEndDate.getDate()-1);//Get end date of this month
        if(now.getDate()>=thisMonthEndDate.getDate()) {//next report period
          reportPeriod = thisMonthEndDate;//Remove period: before end date of this month
        } else {
          reportPeriod = new Date(now.getYear()+1900, now.getMonth(), 1, 0, 0, 0);
          reportPeriod.setDate(reportPeriod.getDate()-1);//Remove period: before end date of last month
        }
      } else {
        if(now.getDate()>=parseInt(result.reportday, 10)){//next report period
          reportPeriod = new Date(now.getYear()+1900, now.getMonth(), parseInt(result.reportday, 10), 0, 0, 0);//Remove period: before $reportday date of this month
        } else {
          reportPeriod = new Date(now.getYear()+1900, now.getMonth()-1, parseInt(result.reportday, 10), 0, 0, 0);//Remove period: before $reportday date of last month
        }
      }
      //console.log("[collectReportSetting] reportPeriod = " + reportPeriod.valueOf() + "," + now.valueOf());
      resolve(reportPeriod.valueOf() + "," + now.valueOf());
    });
  });
}

let promise_action = []
Promise.all([
  dbTplink.waitConnected(),
  dbGenie.waitConnected(),
  dbTmp.waitConnected()
])
.then(() =>{
  if (action==='collect') {
    if (items==='all') {
      promise_action.push(updateCpeInventory(dbGenie, dbTplink, dbTmp))
      promise_action.push(updateUserOperation(dbGenie, dbTplink, dbTmp))
      promise_action.push(updateProfileStats(dbGenie, dbTplink, dbTmp))
      promise_action.push(updateInvalidDevice(dbGenie, dbTplink, dbTmp))
    } else if(items==='reportsetting') {
      promise_action.push(collectReportSetting(dbGenie, dbTplink, dbTmp))
    } else if(items==='testPro'){
      return IsProFeatureSet(dbTplink, function(isPro){
        console.log(isPro);
        process.exit(0);
      });
    }
  } else if (action==='update'){
    if (items==='cpeinventory') {
      promise_action.push(updateCpeInventory(dbGenie, dbTplink, dbTmp))
    } else if (items==='useroperation') {
      promise_action.push(updateUserOperation(dbGenie, dbTplink, dbTmp))
    } else if (items==='profilestats'){
      promise_action.push(updateProfileStats(dbGenie, dbTplink, dbTmp))
    } else if (items==='invaliddevice'){
      promise_action.push(updateInvalidDevice(dbGenie, dbTplink, dbTmp))
    }
  }
  return Promise.all(promise_action)
  .then((result) =>{
    console.log(result)
    // console.log("[" + action +"] ["+items+"]" +" done")
    process.exit(0);
  })
  .catch((err) =>{
    console.log("err = " + err)
    console.log("[" + action +"] ["+items+"]" +" failed")
    process.exit(0);
  })
})
.catch(() =>{
  console.log("Cannot connect to dbTplink & dbGenie & dbTmp");
  process.exit(0);
})
