#!/usr/bin/env node
/*  Copyright(c) 2019-2029 TP-LINK Technologies Co.Ltd.
 *
 * file      decrypt_provisions.js
 * brief     nodejs command to decrypt provisions to redis
 * details   /usr/bin/nodejs /opt/genieacs/bin/decrypt_provisions.js

 * author    Hsinhao Yao
 * version   1.0.0
 * date      10Feb20
 *
 * history   \arg 1.0.0, 10Feb20, Hsinhao Yao, Create the file.
 */

/**************************************************************************************************/
/*                                           VARIABLES                                            */
/**************************************************************************************************/
var fs = require('fs')

const AgileCacheClass = require('tplib/agile-cache')
const AgileCryptor = require('tplib/agile-cryptor')
const MongoClass = require('tplib/mongo-class')

const agileConfig = require('tplib/agile-config')
const stringTable = require('tplib/string-table.json')

const ENCRYPT_ENABLE = agileConfig.get("ENCRYPT_ENABLE")
const LOG_FILE = '/var/log/genieacs/acsapps.log'
const REDIS_FLAG = 'IsDecrypt'
const SYSTEM_LOG_FILE="/var/log/genieacs/acs-system.log"

const enable_debug = false
const enable_logfile = true

const agileCache = new AgileCacheClass(agileConfig.getRedisConfig())
const agileCryptor = new AgileCryptor(ENCRYPT_ENABLE, agileConfig.getRedisConfig())
const dbGenie = new MongoClass(agileConfig.get(stringTable.url.dbGenie))

function endProcess(exitCode) {
  agileCache.disconnect()
  agileCryptor.disconnectRedis()
  dbGenie.disconnectDb()
  process.exit(exitCode)
}

function log(prefix) {
  const current_time = new Date()
  const log_str = "[decrypt_provisions.js] " + current_time + ": " + prefix
  if(enable_debug) console.log(log_str)
  if(enable_logfile) fs.appendFileSync(LOG_FILE, log_str + "\n")
}

Date.prototype.toIsoString = function() {
  function pad(number) {
    if (number < 10) {
      return '0' + number
    }
    return number
  }

  return this.getFullYear() +
    '-' + pad(this.getMonth() + 1) +
    '-' + pad(this.getDate()) +
    'T' + pad(this.getHours()) +
    ':' + pad(this.getMinutes()) +
    ':' + pad(this.getSeconds()) +
    '.' + (this.getMilliseconds() / 1000).toFixed(3).slice(2, 5) +
    'Z'
}

function system_log(prefix) {
  const level = '3'
  const level_str = '[ERROR]'
  const host = '127.0.0.1'
  const objectname = '[Decrypt Provisions]:'
  const current_time = new Date().toIsoString()
  const log_str = current_time + ' ' + level_str + ' ' + level + ' ' + host + ' ' + objectname + ' ' +  prefix
  if(enable_debug) console.log(log_str)
  if(enable_logfile) fs.appendFileSync(SYSTEM_LOG_FILE, log_str + "\n")
}

async function decryptProvisions() {
  if(ENCRYPT_ENABLE) {
    // check if already decrypted
    await agileCache.waitConnected()
    let redisFlag = await agileCache.getPromise(REDIS_FLAG)
    if(redisFlag && redisFlag === '1') return endProcess(0)

    try {
      // wait all db connected
      await Promise.all([
        agileCryptor.waitConnected(),
        dbGenie.waitConnected()
      ])

      // query provision data
      let provisions = await dbGenie.find(stringTable.collection.provisions)
      if(!provisions) {
        log('[decryptProvisions] No encrypt_data')
        return endProcess(0)
      }

      return Promise.all(provisions.map( async provision => {
        log(`[decryptProvisions] ${provision._id} with encrypt == ${provision.encrypt}`)
        if(provision.encrypt === 1) {
          // data is encrypted, decrypt it and set redis
          let result = await agileCryptor.decrypt(provision.script)
          let decrypted = Buffer.from(result, 'base64').toString('ascii')
          await agileCache.setPromise(provision._id, decrypted, -1)
        } else {
          // data is not encrypted, set redis only
          await agileCache.setPromise(provision._id, provision.script)
        }
      }))
      .then(() => {
        log('[Decrypt Provisions] Done')
        agileCache.setPromise(REDIS_FLAG, '1', -1)
        endProcess(0)
      })
      .catch((err) => {
        log('[decryptProvisions] Promise.all err: ' + err)
        endProcess(1)
      })
    } catch(err) {
      system_log('Decrypt provisions failed.')
      log('ERROR:', err)
      endProcess(1)
    }
  } else {
    log('encrypt feature is disabled')
    endProcess(0)
  }
}

decryptProvisions()
