#!/usr/bin/env node
/*  Copyright(c) 2019-2029 TP-LINK Technologies Co.Ltd.
 *
 * file      encrypt_process.js
 * brief     nodejs command to pre-process all encrypt data to mongodb
 * details   /usr/bin/nodejs /opt/genieacs/bin/encrypt_process.js

 * author    Hsinhao Yao
 * version   1.0.0
 * date      12Feb20
 *
 * history   \arg 1.0.0, 12Feb20, Hsinhao Yao, Create the file.
 */

/**************************************************************************************************/
/*                                           VARIABLES                                            */
/**************************************************************************************************/
var fs = require('fs')

const AgileCacheClass = require('tplib/agile-cache')
const AgileCryptor = require('tplib/agile-cryptor')
const MongoClass = require('tplib/mongo-class')

const agileConfig = require('tplib/agile-config')
const stringTable = require('tplib/string-table.json')

const ENCRYPT_DEVICE_ENABLE = false
const ENCRYPT_DEVICE_MAX = 1000
const ENCRYPT_ENABLE = agileConfig.get('ENCRYPT_ENABLE')
const LOG_FILE = '/var/log/genieacs/acsapps.log'
const PROCESS_RUNNING_ID = 'EncryptionIsRunning'

const enable_debug = true
const enable_logfile = true

const agileCache = new AgileCacheClass(agileConfig.getRedisConfig())
const agileCryptor = new AgileCryptor(ENCRYPT_ENABLE, agileConfig.getRedisConfig())
const dbGenie = new MongoClass(agileConfig.get(stringTable.url.dbGenie))
const dbTplink = new MongoClass(agileConfig.get(stringTable.url.dbTplink))

async function endProcess(exitCode) {
  if(exitCode !== 2) {
    // when exitCode is 2 means found duplicated process, do not clean redis flag
    await agileCache.delPromise(PROCESS_RUNNING_ID)
  }
  agileCache.disconnect()
  agileCryptor.disconnectRedis()
  dbGenie.disconnectDb()
  dbTplink.disconnectDb()
  process.exit(exitCode)
}

function log(prefix) {
  const current_time = new Date()
  const log_str = '[encrypt_process.js] ' + current_time + ': ' + prefix
  if(enable_debug) console.log(log_str)
  if(enable_logfile) fs.appendFileSync(LOG_FILE, log_str + "\n")
}

async function encrypt_provisions() {
  let result = await dbGenie.find(stringTable.collection.provisions, {'script': { $exists: true }, 'encrypt': null})
  if(result.length === 0) return
  try {
    await Promise.all(result.map(async r => {
      if(!r || !r.script) throw new Error('encrypted data not found')

      let enc64 = Buffer.from(r['script']).toString('base64')
      let encrypt_result = await agileCryptor.encrypt(enc64)
      await dbGenie.update(stringTable.collection.provisions, { '_id': r['_id']},{ $set: { 'script': encrypt_result, 'encrypt': 1}})
      log(r['_id'] + ' encrypt success!')
    }))
  } catch(err) {
    log(err)
  }
}

async function encrypt_cwmp_settings() {
  try {
    let result = await dbTplink.findOne(stringTable.collection.sysSetting, { '_id': 'cwmp', 'encrypt': null })
    if(!result) return

    let encrypted = await Promise.all([
      agileCryptor.encrypt(result.acs_username).catch(() => result.acs_username),
      agileCryptor.encrypt(result.acs_password).catch(() => result.acs_password),
      agileCryptor.encrypt(result.cpe_username).catch(() => result.cpe_username),
      agileCryptor.encrypt(result.cpe_password).catch(() => result.cpe_password),
    ])
    await dbTplink.update(stringTable.collection.sysSetting, { '_id': 'cwmp' }, {
      $set: {
        'acs_username': encrypted[0],
        'acs_password': encrypted[1],
        'cpe_username': encrypted[2],
        'cpe_password': encrypted[3],
        'encrypt': 1
      }
    })
  } catch (err) {
    log(err)
  }
}

async function encrypt_all_devices () {
  if(!ENCRYPT_DEVICE_ENABLE) return

  let start = new Date().getTime()
  log('Start device encrypt. Time at ' + start)

  let devices = await dbGenie.aggregate(stringTable.collection.devices, [
    { $match : { '_encrypt' : { $ne : 1 }}},
    { $project: {
      'ID' : '$_id',
      'IGDURL': '$InternetGatewayDevice.ManagementServer.URL._value',
      'IGDUsername': '$InternetGatewayDevice.ManagementServer.Username._value',
      'IGDPassword': '$InternetGatewayDevice.ManagementServer.Password._value',
      'IGDConnectionRequestURL': '$InternetGatewayDevice.ManagementServer.ConnectionRequestURL._value',
      'IGPUDPConnectionRequestAddress': '$InternetGatewayDevice.ManagementServer.UDPConnectionRequestAddress._value',
      'IGDConnectionRequestUsername': '$InternetGatewayDevice.ManagementServer.ConnectionRequestUsername._value',
      'IGDConnectionRequestPassword': '$InternetGatewayDevice.ManagementServer.ConnectionRequestPassword._value',
      'DeviceURL': '$Device.ManagementServer.URL._value',
      'DeviceUsername': '$Device.ManagementServer.Username._value',
      'DevicePassword': '$Device.ManagementServer.Password._value',
      'DeviceConnectionRequestURL': '$Device.ManagementServer.ConnectionRequestURL._value',
      'DeviceUDPConnectionRequestAddress': '$Device.ManagementServer.UDPConnectionRequestAddress._value',
      'DeviceConnectionRequestUsername': '$Device.ManagementServer.ConnectionRequestUsername._value',
      'DeviceConnectionRequestPassword': '$Device.ManagementServer.ConnectionRequestPassword._value',
      'XMPPUsername' : '$Device.XMPP.Connection.1.Username._value',
      'XMPPPassword' : '$Device.XMPP.Connection.1.Password._value',
      'XMPPDomain' : '$Device.XMPP.Connection.1.Domain._value',
      'XMPPResource' : '$Device.XMPP.Connection.1.Resource._value',
      'XMPPServerAddress' : '$Device.XMPP.Connection.1.Server.1.ServerAddress._value'
    }}
  ]).toArray()

  if(devices.length === 0) {
    let end = new Date().getTime()
    log('End device encrypt. Time at ' + end)
    log('Used time(Encrypt all devices): ' + (end-start)/1000 + 'sec')
    return
  }

  try {
    await Promise.all(devices.map(async device => {
      if(device.ID) {
        try {
          let match = { '_id': device.ID }
          let update = { '$set': {}}
          let options = {'multi': false}
          if(device.IGDURL) update['$set']['InternetGatewayDevice.ManagementServer.URL._value'] = await agileCryptor.encrypt(device.IGDURL)
          if(device.IGDUsername) update['$set']['InternetGatewayDevice.ManagementServer.Username._value'] = await agileCryptor.encrypt(device.IGDUsername)
          if(device.IGDPassword) update['$set']['InternetGatewayDevice.ManagementServer.Password._value'] = await agileCryptor.encrypt(device.IGDPassword)
          if(device.IGDConnectionRequestURL) update['$set']['InternetGatewayDevice.ManagementServer.ConnectionRequestURL._value'] = await agileCryptor.encrypt(device.IGDConnectionRequestURL)
          if(device.IGPUDPConnectionRequestAddress) update['$set']['InternetGatewayDevice.ManagementServer.UDPConnectionRequestAddress._value'] = await agileCryptor.encrypt(device.IGPUDPConnectionRequestAddress)
          if(device.IGDConnectionRequestUsername) update['$set']['InternetGatewayDevice.ManagementServer.ConnectionRequestUsername._value'] = await agileCryptor.encrypt(device.IGDConnectionRequestUsername)
          if(device.IGDConnectionRequestPassword) update['$set']['InternetGatewayDevice.ManagementServer.ConnectionRequestPassword._value'] = await agileCryptor.encrypt(device.IGDConnectionRequestPassword)
          if(device.DeviceURL) update['$set']['Device.ManagementServer.URL._value'] = await agileCryptor.encrypt(device.DeviceURL)
          if(device.DeviceUsername) update['$set']['Device.ManagementServer.Username._value'] = await agileCryptor.encrypt(device.DeviceUsername)
          if(device.DevicePassword) update['$set']['Device.ManagementServer.Password._value'] = await agileCryptor.encrypt(device.DevicePassword)
          if(device.DeviceConnectionRequestURL) update['$set']['Device.ManagementServer.ConnectionRequestURL._value'] = await agileCryptor.encrypt(device.DeviceConnectionRequestURL)
          if(device.DeviceUDPConnectionRequestAddress) update['$set']['Device.ManagementServer.UDPConnectionRequestAddress._value'] = await agileCryptor.encrypt(device.DeviceUDPConnectionRequestAddress)
          if(device.DeviceConnectionRequestUsername) update['$set']['Device.ManagementServer.ConnectionRequestUsername._value'] = await agileCryptor.encrypt(device.DeviceConnectionRequestUsername)
          if(device.DeviceConnectionRequestPassword) update['$set']['Device.ManagementServer.ConnectionRequestPassword._value'] = await agileCryptor.encrypt(device.DeviceConnectionRequestPassword)
          if(device.XMPPUsername) update['$set']['Device.XMPP.Connection.1.Username._value'] = await agileCryptor.encrypt(device.XMPPUsername)
          if(device.XMPPPassword) update['$set']['Device.XMPP.Connection.1.Password._value'] = await agileCryptor.encrypt(device.XMPPPassword)
          if(device.XMPPDomain) update['$set']['Device.XMPP.Connection.1.Domain._value'] = await agileCryptor.encrypt(device.XMPPDomain)
          if(device.XMPPResource) update['$set']['Device.XMPP.Connection.1.Resource._value'] = await agileCryptor.encrypt(device.XMPPResource)
          if(device.XMPPServerAddress) update['$set']['Device.XMPP.Connection.1.Server.1.ServerAddress._value'] = await agileCryptor.encrypt(device.XMPPServerAddress)
          update['$set']['_encrypt'] = 1
          await dbGenie.update(stringTable.collection.devices, match, update, options)
        } catch(err) {
          log(err)
        }
      }
    }))
    let end = new Date().getTime()
    log('Used time(Encrypt ' + ENCRYPT_DEVICE_MAX/1000 + 'K devices): ' + (end - start)/1000 + 'sec')
  } catch(err) {
    log(err)
  }
}

async function encrypt_notification() {
  let mails = await dbTplink.aggregate(stringTable.collection.notification, [
    { $match : { '_id': 'mail' }}, {
      $project: {
        'username' : '$settings.username',
        'password' : '$settings.password',
        'encrypt' : '$settings.encrypt'
      }
    }
  ]).toArray()
  if(mails.length === 0) return

  if(!mails[0].encrypt || mails[0].encrypt !== 1) {
    try {
      let mail_username = mails[0].username ? await agileCryptor.encrypt(mails[0].username) : ''
      let mail_password = mails[0].password ? await agileCryptor.encrypt(mails[0].password) : ''
      await dbTplink.update(stringTable.collection.notification, { '_id': mails[0]._id }, { $set: {
          'settings.username': mail_username,
          'settings.password': mail_password,
          'settings.encrypt': 1
        }
      })
    } catch(err) {
      log(err)
    }
  }
}

async function encryptProcess() {
  // check if encrypt is enabled
  if(!ENCRYPT_ENABLE) {
    log('encrypt feature is disabled')
    return endProcess(1)
  }

  // check if other encrypt process running, redis way
  await agileCache.waitConnected()
  let backupTtl = await agileCache.ttlPromise(PROCESS_RUNNING_ID)
  let isRunning = await agileCache.getSetPromise(PROCESS_RUNNING_ID, '1')
  if(isRunning === '1') {
    log('duplicated encryption process found')
    // since redis.getset() will make key permernant, so try to restore the original ttl
    if(backupTtl > 0) await agileCache.expirePromise(PROCESS_RUNNING_ID, backupTtl)
    return endProcess(2)
  }

  // no other process is running, set a default ttl
  agileCache.expire(PROCESS_RUNNING_ID, 600)

  // wait dbs to be connected
  try {
    await Promise.all([
      agileCryptor.waitConnected(),
      dbGenie.waitConnected(),
      dbTplink.waitConnected()
    ])

    await Promise.all([
      encrypt_provisions(),
      encrypt_cwmp_settings(),
      encrypt_all_devices(),
      encrypt_notification()
    ])
    endProcess(0)
  } catch(err) {
    log(err)
    endProcess(1)
  }
}

encryptProcess()
