/*  Copyright(c) 2020-2025 TP-LINK Technologies Co.Ltd.
 *
 * file      las-api.js
 * brief     HTTPS restful API for update ativate report to LAS

 * author    Dean Hsiao
 * version   1.0.0
 * date      23 Apr 2020
 *
 * history   \arg 1.0.0, 23 Apr 2020, Dean Hsiao, Create the file.
 */

var https	= require('https');
var fs		= require('fs');


var HTTPS_MAX_RETRY	= 3;
var HTTPS_TIMEOUT	= 5000;
var HTTPS_BEARER_KEY		= "/opt/agileacs/data/las_api/Bearer_key";
var HTTPS_NEW_BEARER_KEY	= "/opt/agileacs/data/las_api/Bearer_new_key";

const LAS_API_DEBUG_FILE="/var/log/genieacs/las_api.log";
var las_api_debug_enable = true;

let log_debug = function (msg)
{
	if(!las_api_debug_enable)
		return;
	let current_time = new Date().toIsoString();
	let log_str = current_time + ' ' +  msg;
	fs.appendFileSync(LAS_API_DEBUG_FILE, log_str+"\n");
	//console.log(log_str);
}

function readFileContent(filename) {
	return fs.readFileSync(filename, 'utf8');
}

function writeFileContent(filename, content) {
	return fs.writeFileSync(filename, content, 'utf8');
}

function getBearerKey() {
	if(fs.existsSync(HTTPS_NEW_BEARER_KEY)) {
		return readFileContent(HTTPS_NEW_BEARER_KEY).trim();
	} else {
		return readFileContent(HTTPS_BEARER_KEY).trim();
	}
}

function delay(s) {
	return new Promise(function(resolve, reject){
		setTimeout(resolve, s);
	});
}

function sendDataToURL(host, path, json_data) {
	return new Promise((resolve, reject) => {
		var bearer_key = getBearerKey();
		var result = {};

		var options = {
			//hostname: '192.168.20.234',
			//path: '/api/v1/agileacs_update',
			hostname: host,
			path: path,
			port: 443,
			method: 'POST',
			headers: {
				'Authorization': 'Bearer ' + bearer_key,
				'Content-Type': 'application/json',
				'Content-Length': json_data.length
			},
			rejectUnauthorized: false,	// ignore self signed certificate
		};

		const req = https.request(options, res => {
			switch(res.statusCode) {
			case 200:
				res.on('data', d => {
					if(json_ret = JSON.parse(d.toString())) {
						if(json_ret.error) {
							result['status'] = '400';
							result['message'] = 'Bad Request';
							reject(result);
						} else {
							result['status'] = res.statusCode + '';
							result['message'] = res.statusMessage;
							resolve(result);
						}
					} else {
						result['status'] = res.statusCode + '';
						result['message'] = res.statusMessage;
						resolve(result);
					}
				});
				break;
			default:
				result['status'] = res.statusCode + '';
				result['message'] = res.statusMessage;
				result['err'] = res.statusMessage;
				reject(result);
				break;
			}
		});

		req.on('socket', function (socket) {
			socket.setTimeout(HTTPS_TIMEOUT);
			socket.on('timeout', function() {
				result['status'] = '408';
				result['message'] = 'Request Timeout';
				result['err'] = 'Request Timeout';
				reject(result);

				req.abort();
			});
		});

		req.on('error', error => {
			result['status'] = '503';
			result['message'] = 'Service Unavailable';
			result['err'] = 'Service Unavailable';
			reject(result);
		});

		req.on('timeout', res => {
			result['status'] = '408';
			result['message'] = 'Request Timeout';
			result['err'] = 'Request Timeout';
			reject(result);
		});

		req.write(json_data);
		req.end();
	});
};


function updateApiKey(host) {
	json_data = '{"action":"get_key"}';
	path = '/agileacs_get_key';

	var result = {};

	return new Promise((resolve, reject) => {
		//var bearer_key = readFileContent('Bearer_key.txt').trim();

		var options = {
			hostname: host,
			path: path,
			port: 443,
			method: 'GET',
			headers: {
				//'Authorization': 'Bearer ' + bearer_key,
				'Content-Type': 'application/json',
				'Content-Length': json_data.length
			},
			rejectUnauthorized: false,	// ignore self signed certificate
		};

		const req = https.request(options, res => {
			switch(res.statusCode) {
			case 200:
				res.on('data', d => {
					json_res = JSON.parse(d.toString());
					writeFileContent(HTTPS_NEW_BEARER_KEY, json_res.api_key.trim());

					result['status'] = res.statusCode + '';
					result['message'] = res.statusMessage;
					resolve(result);
				});
				break;
			default:
				result['status'] = res.statusCode + '';
				result['message'] = res.statusMessage;
				result['err'] = res.statusMessage;
				resolve(result);
				break;
			}
		});

		req.on('socket', function (socket) {
			socket.setTimeout(HTTPS_TIMEOUT);
			socket.on('timeout', function() {
				req.abort();
			});
		});

		req.on('error', error => {
			result['status'] = '503';
			result['message'] = 'Service Unavailable';
			result['err'] = 'Service Unavailable';
			reject(result);
		});

		req.on('timeout', res => {
			result['status'] = '408';
			result['message'] = 'Request Timeout';
			result['err'] = 'Request Timeout';
			reject(result);
		});

		req.write(json_data);
		req.end();
	});
}

async function sendToLAS(url, json_data, retry = 0) {
	let retryCount = retry
	let url_arr = url.split('/'), path = ''
	let host = url_arr[2]
	for(let i=3; i<url_arr.length; i++) {
		path += "/" + url_arr[i]
	}

	while(retryCount < HTTPS_MAX_RETRY) {
		try {
			return await sendDataToURL(host, path, JSON.stringify(json_data))
		} catch(err) {
			if(err['status'] == '400') return err
			else {
				if(err['status'] == '302') {
					log_debug('API Key fail (updatekey)')
					await updateKey(url, 1)
				}
				await delay(3000)
				log_debug('sendToLAS() retry [status: ' + err['status'] + ', message: ' + err['message'] + ']')
				retryCount++
			}
		}
	}

	// too many retries
	return ({
		status: '503',
		message: 'Service Unavailable (exceeded max retry)',
		err: 'Service Unavailable (exceeded max retry)'
	})
}


function updateKey(url, retry) {
	return new Promise((resolve) => {
		var result = {};

		if(retry > HTTPS_MAX_RETRY) {
			result['status'] = '503';
			result['message'] = 'Service Unavailable (exceeded max retry)';
			result['err'] = 'Service Unavailable (exceeded max retry)';
			return resolve(result);
		}

		url_arr = url.split('/');
		path = "";

		host = url_arr[2];
		for(i=3; i<url_arr.length; i++) {
			path += "/" + url_arr[i];
		}

		return updateApiKey(host)
		.then((res) => {
			return resolve(res);
		})
		.catch((err) => {
			delay(3000).then(function() {
				log_debug('updateKey() retry [status: ' + err['status'] + ', message: ' + err['message'] + ']');
				return resolve(updateKey(url, retry+1));
			});
		});
	});
}

exports.sendToLAS = sendToLAS;
exports.updateKey = updateKey;
