#!/usr/bin/env node
"use strict"
const ACS_SYSTEM_LOG_FILE="/var/log/genieacs/acs-system.log"
const nodemailer = require('nodemailer')
const fs = require('fs')
const { agileConfig, stringTable, MongoClass, AgileCryptor } = require('tplib')
const args = process.argv.slice(2)
const event_id = args[0]

const tpDb = new MongoClass(agileConfig.get(stringTable.url.dbTplink))
const ac = new AgileCryptor(agileConfig.get('ENCRYPT_ENABLE'), agileConfig.getRedisConfig())

const ENCRYPT_ENABLE = agileConfig.get('ENCRYPT_ENABLE')
const SENDMAIL_TIMEOUT = 30000

Date.prototype.toIsoString = function() {
  function pad(number) {
    if (number < 10) {
      return '0' + number
    }
    return number
  }
  return this.getFullYear() +
    '-' + pad(this.getMonth() + 1) +
    '-' + pad(this.getDate()) +
    'T' + pad(this.getHours()) +
    ':' + pad(this.getMinutes()) +
    ':' + pad(this.getSeconds()) +
    '.' + (this.getMilliseconds() / 1000).toFixed(3).slice(2, 5) +
    'Z'
}

function logger(msg, level_str) {
  let level = get_level(level_str)
  let host = '127.0.0.1'
  let objectname = '[Notifications]:'
  let current_time = new Date().toIsoString()
  let log_str = current_time + ' [' + level_str.toUpperCase() + '] ' + level + ' ' + host + ' ' + objectname + ' ' +  msg

  fs.appendFileSync(ACS_SYSTEM_LOG_FILE, log_str+"\n")
  return
}

function get_level(level_str) {
  let severity = "0"
  if (level_str.match(/debug/i))
    severity = "7"
  else if (level_str.match(/info/i))
    severity = "6"
  else if (level_str.match(/notice/i))
    severity = "5"
  else if (level_str.match(/warn/i))
    severity = "4"
  else if (level_str.match(/error/i))
    severity = "3"
  else if (level_str.match(/critical/i))
    severity = "2"
  else if (level_str.match(/alert/i))
    severity = "1"
  else if (level_str.match(/emergency/i))
    severity = "0"

  return severity
}

function transfer_to_json(data) {
  const data_to_string = data.replace(/\//g, '"')
  const data_json = JSON.parse(data_to_string)
  return data_json
}

async function get_event_template(id){
  try {
    await tpDb.waitConnected()
    let result = await tpDb.find(stringTable.collection.notification, { "_id": "mail" }, { "event_list": 1 })
    return {
      subject: result[0]["event_list"][id]["subject"],
      context: result[0]["event_list"][id]["context"]
    }
  } catch(e) {
    console.log(e)
  }
}

async function get_common_template(){
  try {
    await tpDb.waitConnected()
    let result = await tpDb.find(stringTable.collection.notification, { "_id": "common" })
    return result[0]["template"]
  } catch(e) {
    console.log(e)
  }
}

async function get_mail_settings(){
  try {
    await tpDb.waitConnected()
    let result = await tpDb.find(stringTable.collection.notification, { "_id": "mail" }, { "settings": 1 })
    return result[0]["settings"]
  } catch(e) {
    console.log(e)
  }
}

async function generate_mail_configs(mail, event_id) {
  let configs = {}
  switch(mail.service) {
    case "gmail":
      configs["service"] = "gmail"
      break
    case "custom":
      configs = { host: mail.smtp_server, port: mail.port }
      break
    case "outlook":
      configs["service"] = "outlook"
      break
    default:
      console.log("No selected services.")
  }

  if(event_id != "000" && ENCRYPT_ENABLE) {
    mail.username = await ac.decrypt(mail.username)
    mail.password = await ac.decrypt(mail.password)
  }
  if(mail.enable_auth == "true")
    configs["auth"] = { user: mail.username, pass: mail.password }

  return configs
}

function generate_mail_options(mail, template, common) {
    let date = new Date()
    const date_s = date.toLocaleDateString().split("/")
    const sign_date = date_s[2] + "/" + date_s[0] + "/" + date_s[1] + " " + date.toTimeString().replace(/ \(.+\)/,"")
    let options = { from: mail.from, to: mail.to }
    if (mail.subject_prefix.length > 0)
      options["subject"] = "[" + mail.subject_prefix + "]" + template.subject
    else
      options["subject"] = template.subject

    options["text"] = common.greeting + '\n\n\t' + template.context + '\n\n\n' + common
    .sign
    .replace("%time%", sign_date)
    return options
}

async function main(){
  const template = await get_event_template(event_id)
  const common = await get_common_template()
  // console.log(template, common)
  let mail = {}
  if (event_id != "000"){
    mail = await get_mail_settings()
  }
  else{
    const data = args[1]
    mail = transfer_to_json(data)
  }

  const mailConfigs = await generate_mail_configs(mail, event_id)
  const mailOptions = generate_mail_options(mail,template, common)
  const transport = nodemailer.createTransport(mailConfigs)

  let timeoutProtect = setTimeout(function() {
    timeoutProtect = null
    let msg = "Error:Timeout" + ' Type:' + event_id
    console.log(msg)
    logger(msg, "error")
    process.exit()
  }, SENDMAIL_TIMEOUT)

  return transport.sendMail(mailOptions, function(error, info) {
    let msg = ""
    if (timeoutProtect)
      clearTimeout(timeoutProtect)

    if (error) {
      if (error.responseCode == "535")
        msg = "Authentication unsuccessful " + error.responseCode + ' Type:' + event_id
      else if (error.responseCode == "530")
        msg = "Client was not authenticated "+ error.responseCode + ' Type:' + event_id
      else
        msg = "Unknown Error: " + error + ' Type:' + event_id

      console.log(msg)
      logger(msg, "error")
    } else {
      msg = 'Email sent: ' + info.response + ' Type:' + event_id
      console.log(msg)
      logger(msg, "info")
    }
    tpDb.disconnectDb()
    ac.disconnectRedis()
    process.exit()
  })
}
main()
