/*
 * Copyright 2019-2029 Agile ACS
 *
 * This file is part of Agile ACS
 *   This file is a part of Agile NBI
 */
"use strict";

const cluster = require("cluster");
const logger = require("tplib").agileLogger;
const agileConfig = require('tplib').agileConfig

const VERSION = agileConfig.get("VERSION");

let respawnTimestamp = 0;
let tooManyCrashesTimestamp = 0;

function restartWorker(worker, code, signal) {
  let msg = {
    message: "Worker died",
    pid: worker.process.pid
  };

  if (code != null)
    msg.exitCode = code;

  if (signal != null)
    msg.signal = signal;

  logger.error(msg);

  const now = Date.now();
  respawnTimestamp = Math.max(now, respawnTimestamp + 2000);
  if (respawnTimestamp === now) {
    tooManyCrashesTimestamp = now;
    cluster.fork();
    return
  }

  if (now - tooManyCrashesTimestamp > 60000) {
    process.exitCode = 1;
    cluster.removeListener("exit", restartWorker);
    for (let pid in cluster.workers)
      cluster.workers[pid].kill();

    logger.error({
      message: "Too many crashes, exiting",
      pid: process.pid
    });
    return;
  }

  setTimeout(function() {
    if (process.exitCode) return;
    cluster.fork();
  }, respawnTimestamp - now);
}

function start(service) {
  try {
    logger.init(service, VERSION);

    const SERVICE_ADDRESS = agileConfig.get(service+"_INTERFACE", true);
    const SERVICE_PORT = agileConfig.get(service+"_PORT", true);

    logger.info({
      message: `AgileACS (${service}) starting`,
      pid: process.pid,
      version: VERSION
    });

    cluster.on("listening", function(worker, address) {
      if ((address.addressType === 4 || address.addressType === 6) &&
          address.address === SERVICE_ADDRESS && address.port === SERVICE_PORT) {
        logger.info({
          message: "Worker listening",
          pid: worker.process.pid,
          address: address.address,
          port: address.port
        });
      }
    });

    cluster.on("exit", restartWorker);

    process.on("SIGINT", function() {
      logger.info({
        message: "Received signal SIGINT, exiting",
        pid: process.pid
      });
      logger.disconnectDb()
      logger.close()

      cluster.removeListener("exit", restartWorker);
    });

    process.on("SIGTERM", function() {
      logger.info({
        message: "Received signal SIGTERM, exiting",
        pid: process.pid
      });
      logger.disconnectDb()
      logger.close()

      cluster.removeListener("exit", restartWorker);
      for (let pid in cluster.workers)
        cluster.workers[pid].kill();
    });

    cluster.setupMaster({
      exec: "lib/modularization/modularization-server",
      args: [service]
    });

    let workerCount = agileConfig.get(`${service}_WORKER_PROCESSES`, true);

    if (!workerCount)
      workerCount = Math.max(2, require("os").cpus().length);

    for (let i = 0; i < workerCount; ++ i)
      cluster.fork();
  } catch(ex) {
    console.log("GGG: catch ex=", ex);
  }
}


process.on("uncaughtException", function(err) {
  logger.error({
    message: "Uncaught exception in master process, exiting",
    exception: err,
    pid: process.pid
  });
  process.exitCode = 1;
  cluster.removeListener("exit", restartWorker);
  for (let pid in cluster.workers)
    cluster.workers[pid].kill();
});


exports.start = start;
