'use strict'

const CACHE_TIMEOUT = 500

class AgileCacheClass {
  constructor(config, callback) {
    this.config = config
    this.dbCli = null
    this.isConnected = false

    this.connectCache(callback)
  }

  connectCache(callback) {
    let connect = () => {
      if(this.config.redisHost) {
        this.dbCli = require("redis").createClient(this.config.redisPort, this.config.redisHost)
        this.dbCli.select(this.config.redisDb, (err, value) => {
          if(callback) callback(err, value)
        })

        this.dbCli.on("error", err => {
          switch(err.code) {
            case 'ECONNREFUSED':
              break
            default:
              console.log(err.code)
          }
        })

        this.dbCli.on("connect", () => {
          this.isConnected = true
        })
      }
    }

    connect()
  }

  async waitConnected() {
    return new Promise(res => {
      let check = () => {
        if(this.isConnected) {
          return res()
        }
        setTimeout(check, 200)
      }
      check()
    })
  }

  disconnect() {
    this.isConnected = false
    if(this.dbCli) {
      if(this.config.redisHost) this.dbCli.quit()
      else this.dbCli.close()  // for mongo
    }
  }

  get(key, callback) {
    if(this.config.redisHost) {
      if (Array.isArray(key)) this.dbCli.mget(key, callback)
      else this.dbCli.get(key, callback)
    }
  }

  // will resolve with null if key not exists
  getPromise(key) {
    return new Promise((res, rej) => {
      let timer = setTimeout(() => rej(), CACHE_TIMEOUT)

      this.get(key, (err, result) => {
        clearTimeout(timer)
        if(err) return rej(err)
        res(result)
      })
    })
  }

  set(key, value, ttl, callback) {
    let ttlLocal = ttl
    let cb = callback

    // check third parameter
    if (!callback && typeof ttl === "function") {
      cb = ttl
      ttlLocal = null
    }

    // check ttl range
    if(!ttlLocal || ttlLocal < 0) ttlLocal = this.config.maxCacheTtl

    if(this.config.redisHost) {
      this.dbCli.setex(key, ttlLocal, value, function(err, result){
        if(ttl === -1){
          this.dbCli.persist(key)
        }
        if(cb) cb(err, result)
      }.bind(this))
    }
  }

  setPromise(key, value, ttl) {
    return new Promise((res, rej) => {
      let timer = setTimeout(() => rej(), CACHE_TIMEOUT)

      this.set(key, value, ttl, (err, result) => {
        clearTimeout(timer)
        if(err) return rej(err)
        res(result)
      })
    })
  }

  getSetPromise(key, value) {
    return new Promise((res, rej) => {
      let timer = setTimeout(() => rej(), CACHE_TIMEOUT)

      this.dbCli.getset(key, value, (err, result) => {
        clearTimeout(timer)
        if(err) return rej(err)
        res(result)
      })
    })
  }

  del(key, callback) {
    if(this.config.redisHost) this.dbCli.del(key, callback)
  }

  // will resolve with 0 if key not exists
  delPromise(key) {
    return new Promise((res, rej) => {
      let timer = setTimeout(() => rej(), CACHE_TIMEOUT)

      this.del(key, (err, result) => {
        clearTimeout(timer)
        if(err) return rej(err)
        res(result)
      })
    })
  }

  pop(key, callback) {
    const script = "local v=redis.call(\"get\",KEYS[1]);redis.call(\"del\",KEYS[1]);return v;"
    if(this.config.redisHost) this.dbCli.eval(script, 1, key, callback)
  }

  lock(lockName, ttl, callback) {
    if(this.config.redisHost) {
      const UNLOCK_SCRIPT = "if redis.call(\"get\", KEYS[1]) == ARGV[1] then return redis.call(\"del\", KEYS[1]) else return 0 end"
      const EXTEND_SCRIPT = "if redis.call(\"get\", KEYS[1]) == ARGV[1] then return redis.call(\"pexpire\", KEYS[1], ARGV[2]) else return 0 end"
      const token = Math.random().toString(36).slice(2)
      const redisClient = this.dbCli

      function unlockOrExtend(ttl) {
        if (!ttl) {
          redisClient.eval(UNLOCK_SCRIPT, 1, lockName, token, function(err, res) {
            if (err || !res)
              throw err || new Error("Lock expired")
          })
        }
        else {
          redisClient.eval(EXTEND_SCRIPT, 1, lockName, token, ttl, function(err, res) {
            if (err || !res)
              throw err || new Error("Lock expired")
          })
        }
      }

      redisClient.set(lockName, token, "NX", "EX", ttl, function(err, res) {
        if (err || res) {
          if(callback) return callback(err, unlockOrExtend)
          else return
        }

        setTimeout(function() {
          this.lock(lockName, ttl, callback)
        }.bind(this), 200)
      }.bind(this))
    }
  }

  ttl(key, callback) {
    if(this.config.redisHost) this.dbCli.ttl(key, callback)
  }

  // will resolve with -2 if key not exists
  ttlPromise(key) {
    return new Promise((res, rej) => {
      let timer = setTimeout(() => rej(), CACHE_TIMEOUT)
      if(!this.config.redisHost) return rej()
      this.ttl(key, (err, ttl) => {
        clearTimeout(timer)
        if(err) return rej()
        res(ttl)
      })
    })
  }

  expire(key, ttl, callback) {
    if(this.config.redisHost) this.dbCli.expire(key, ttl, callback)
  }

  expirePromise(key, ttl) {
    return new Promise((res, rej) => {
      let timer = setTimeout(() => rej(), CACHE_TIMEOUT)
      if(!this.config.redisHost) return rej()
      this.expire(key, ttl, (err, result) => {
        clearTimeout(timer)
        if(err) return rej()
        res(result)
      })
    })
  }
}

module.exports = AgileCacheClass
