'use strict'

const AgileCacheClass = require('./agile-cache')
const AgileACSApps = require('./agile-acsapps')

class AgileCryptor {
  constructor(encryptEnabled, cacheConfig) {
    this.encryptEnabled = encryptEnabled
    this.cache = new AgileCacheClass(cacheConfig)
  }

  get_key() {
    let getKeyFromSystem = () => {
      return AgileACSApps.execCmdWithSafeArgs("/opt/agileacs/bin/acsutils2", "getserverinfo")
    }

    return this.cache.getPromise("Encrypt_key")
    .then(redisCached => {
      if(!redisCached || redisCached === "") return getKeyFromSystem().then(result => {
        // set cache but not block the encrypt/decrypt process
        this.cache.setPromise("Encrypt_key", result.stdout, -1)
        return result.stdout
      })
      return redisCached
    })
    .catch(() => {
      // redis timeout
      // so we do not try to update redis cache here
      return getKeyFromSystem().then(result => result.stdout)
    })
  }

  gen_key(key) {
    return new Promise((resolve, reject) => {
      if(!key || key == "") {
        this.get_key()
        .then(result => resolve(result))
        .catch((e) => reject(e))
      }
      else resolve(key)
    })
  }

  /**
   * Update System Key to Redis
   */
  async updateKeyCache() {
    try {
      let { stdout } = await AgileACSApps.execCmdWithSafeArgs("/opt/agileacs/bin/acsutils2", "getserverinfo")
      await this.cache.setPromise("Encrypt_key", stdout, -1)
    } catch(err) {
      console.log('get key from system failed')
    }
  }

  encrypt(input, key) {  // key format is "cpu-id" "hdd-id"
    return new Promise((resolve, reject) => {
      if(!this.encryptEnabled || !input || input === "") resolve(input)
      else {
        this.gen_key(key).then(legalKey => {
          let saftInput = input.replace(/\\/g, '\\\\').replace(/"/g, '\\"')
          AgileACSApps.exec(`/opt/agileacs/bin/acsutils2 encrypt "${saftInput}" ${legalKey}`, function(err,stdout,stderr){
            if(err) reject(err)
            if(stderr) reject(stderr)
            resolve(stdout)
          })
        })
        .catch((e) => reject(e))
      }
    })
  }

  /**
   * Decrypt String
   * @param {string} input encrypted string to be decrypted
   * @param {string} key optional: decrypt key, no need to provide this if we want to use system key
   * @returns {Promise<string>} decrypted string. If error occurs when decrypt, return original string
   */
  decrypt(input, key) {
    return new Promise((resolve) => {
      if(!this.encryptEnabled || !input || input === "") resolve(input)
      else {
        this.gen_key(key).then(legalKey => {
          let saftInput = input.replace(/\\/g, '\\\\').replace(/"/g, '\\"')
          AgileACSApps.exec(`/opt/agileacs/bin/acsutils2 decrypt "${saftInput}" ${legalKey}`, function(err, stdout, stderr){
            if(err || stderr) resolve(input)
            resolve(stdout)
          })
        })
        .catch(() => resolve(input))
      }
    })
  }

  async waitConnected() {
    return this.cache.waitConnected()
  }

  disconnectRedis() {
    this.cache.disconnect()
  }
}

module.exports = AgileCryptor
