'use strict'

const http = require('http')

const AgileCacheClass = require('tplib/agile-cache')
const MongoClass = require('tplib/mongo-class')
const agileConfig = require('tplib/agile-config')
const stringTable = require('tplib/string-table.json')

const HTTP_REQUEST_TIMEOUT = 5000

const dbGenie = new MongoClass(agileConfig.get(stringTable.url.dbGenie))
const agileCache = new AgileCacheClass(agileConfig.getRedisConfig())
const diag_timers = {}

var find_value = function(target, key) {
  let ret = null, keys = Object.keys(target)
  if( keys.indexOf(key) >= 0) ret = target[key]
  return ret
}

function extract_value(device, path) {
  let ret
  if ( !device || !path ) return ret

  let tokens = path.split('.')
  if (tokens.length > 0) {
    var ref1 = device, ref2 = device
    for(var i = 0; i < tokens.length; i++) {
      var token = tokens[i]
      if (!token.startsWith('_')) {
        ref2 = find_value(ref1, token)

        if (ref2) ref1 = ref2
        else break
      }
    }
    ret = ref1['_value']
  }

  return ret
}

function send_task(deviceId, task, append_url='') {
  return new Promise((resolve, reject)=>{
    let json_data = JSON.stringify(task)
    var http_options = {
      hostname: agileConfig.get('AGILE_NBI_INTERFACE'),
      port: agileConfig.get('AGILE_NBI_PORT'),
      path: '/devices/' + encodeURI(deviceId) + '/tasks?timeout=3000&connection_request' + append_url,
      method: 'POST',
      headers: {
        'Content-Type': 'application/json',
        'Content-Length': Buffer.byteLength(json_data),
        'Connection': 'close'
      }
    }
    var req = http.request(http_options, res => {
      let data
      res.on('data', function (chunk) {
        data += chunk
      })
      res.on('end', function () {
        return resolve(data)
      })
      res.on('error', error => {
        return reject(error)
      })
      res.on('socket', function (socket) {
        socket.setTimeout(HTTP_REQUEST_TIMEOUT)
        return socket.on('timeout', () => {
          if (!req.aborted) {
            req.abort()
          }
          return reject('timeout')
        })
      })
    })
    req.setTimeout(5000, () => {
      if (!req.aborted) {
        req.abort()
        return reject('timeout')
      }
    })
    req.on('error', (error) => {
      if (!req.aborted) {
        req.abort()
        return reject(error)
      }
    })
    req.write(json_data)
    req.end()
  })
}

function disable_UDPEchoConfig(_id) {
  return dbGenie.findOne(stringTable.collection.devices, { _id }, {
    "_id": 1,
    "_deviceId": 1,
    "InternetGatewayDevice.UDPEchoConfig.Enable._value": 1,
    "Device.IP.Diagnostics.UDPEchoConfig.Enable._value": 1
  })
  .then(result => {
    if(!result) return remove_diag_timer(_id)

    let path, value
    let paths = ['InternetGatewayDevice.UDPEchoConfig.Enable', 'Device.IP.Diagnostics.UDPEchoConfig.Enable']
    for(let i=0; i<paths.length; i++) {
      path = paths[i]
      value = extract_value(result, path)
      if(value) break
    }
    if(value) {
      let task = { 'name': 'setParameterValues', 'parameterValues': [[path, false]] }
      return send_task(_id, task)
      .then(() => {
        task = { 'name': 'refreshObject', 'objectName': path.substr(0, path.indexOf('.Enable')) }
        return send_task(_id, task)
      })
      .catch(err => {
        console.log(`err:${err}`)
      })
    }
  })
  .catch(err => {
    console.log(err)
  })
}

function stop_diag(deviceId, diagName) {
  if(diagName === 'UDPEchoConfig') {
    return new Promise((res, rej) => {
      agileCache.getPromise(`${deviceId}_diaglock`)
      .then(() => {
        return disable_UDPEchoConfig(deviceId)
        .then(() => {
          agileCache.delPromise(`${deviceId}_diaglock`)
          res(deviceId)
        })
      })
      .catch(err => {
        rej(err)
      })
    })
  }
  return Promise.resolve(deviceId)
}

function insert_diag_timer(deviceId, diagName, ttl) {
  let timeout = Number.parseInt(ttl) * 1000
  try {
    if(!deviceId || deviceId === '' || diagName !== 'UDPEchoConfig' || ttl < 1) return false

    remove_diag_timer(deviceId)
    diag_timers[deviceId] = setTimeout(() => {
      stop_diag(deviceId, diagName)
      .catch(err => {
        console.log(`stop_diag: error, ${err}`)
      })
      remove_diag_timer(deviceId)
    }, timeout)
  } catch(err) {
    console.log(err)
  }
}

function remove_diag_timer(deviceId) {
  try {
    if(!deviceId || deviceId === '') return false
    if(diag_timers[deviceId]) {
      clearTimeout(diag_timers[deviceId])
      delete diag_timers[deviceId]
    }
  } catch(err) {
    console.log(err)
  }
  return true
}

function clear() {
  try {
    Object.keys(diag_timers).forEach(key => {
      remove_diag_timer(key)
    })
  } catch(err) {
    console.log(err)
  }
}

function disconnectDB() {
  agileCache.disconnect()
  dbGenie.disconnectDb()
}

module.exports = {
  insert_diag_timer,
  remove_diag_timer,
  clear,
  disconnectDB
}
