'use strict'

let MongoClient = require('mongodb').MongoClient
let stringTable = require('./string-table.json')

// default settings, can not be modified
const defaultSettings = {
  timeout: 5000
}
Object.freeze(defaultSettings)

/**
 * Class that implements API to access mongodb.
 * Using mongodb driver 2.2
 * @param {string} URL mongodb url to connect
 * @param {object} settings settings for this db client
 *
 */
class MongoClass {
  constructor(URL, settings) {
    this.db = null
    this.url = URL
    this.timer = null
    this.isConnecting = false
    this.settings = Object.assign({}, defaultSettings, settings)
    this.connectDB(URL)
  }

  connectDB(URL) {
    this.isConnecting = true
    MongoClient.connect(URL, { reconnectTries: 0 })
    .then(db => {
      // console.log(stringTable.message.connectDBSuccess + ` (${URL})`)
      this.db = db

      db.on('close', () => {
        this.db.removeAllListeners()
        this.db = null
        this.reconnectDB()
      })
    })
    .catch(() => {
      console.log(stringTable.error.dbNotConnected + ` (${URL})`)
      this.timer = setTimeout(() => {
        this.timer = null
        this.reconnectDB()
      }, 5000)
    }).then(() =>{
      this.isConnecting = false
    })
  }

  reconnectDB() {
    this.connectDB(this.url)
  }

  disconnectDb() {
    if(this.timer) clearTimeout(this.timer)
    if (this.db) {
      this.db.removeAllListeners()
      this.db.close()
      this.db = null
    }
    return new Promise(res => {
      let checkConnecting = () => {
        if(!this.isConnecting) return res()
        setTimeout(checkConnecting, 1000)
      }
      checkConnecting()
    })
  }

  isConnected() {
    if (this.db) return true
    return false
  }

  async waitConnected() {
    let retryTimes = 0
    let retryFunction = (res, rej) => {
      if(this.isConnected()) return res('connected')
      if(retryTimes > 3) return rej('db is down?')
      retryTimes++
      setTimeout(() => {
        retryFunction(res, rej)
      }, 1000 * retryTimes)
    }
    return new Promise((res, rej) => {
      retryFunction(res, rej)
    })
  }

  handleError(callback) {
    if (callback) return callback(new Error(stringTable.error.dbNotConnected + ` (${this.url})`))
    return Promise.reject(new Error(stringTable.error.dbNotConnected + ` (${this.url})`))
  }

  find(collection, ...args) {
    let callback = null
    if(args.length >= 1 && typeof args[args.length-1] === 'function') {
      callback = args.pop()
    }
    if (!this.isConnected()) return this.handleError(callback)

    if(!callback) return this.db.collection(collection).find(...args).toArray()
    return this.db.collection(collection).find(...args).toArray(callback)
  }

  findOne(collection, ...args) {
    let callback = null
    if(args.length >= 1 && typeof args[args.length-1] === 'function') {
      callback = args[args.length-1]
    }
    if (!this.isConnected()) return this.handleError(callback)

    return this.db.collection(collection).findOne(...args)
  }

  insert(collection, ...args) {
    let callback = null
    if(args.length >= 1 && typeof args[args.length-1] === 'function') {
      callback = args[args.length-1]
    }
    if (!this.isConnected()) return this.handleError(callback)

    return this.db.collection(collection).insert(...args)
  }

  insertOne(collection, ...args) {
    let callback = null
    if(args.length >= 1 && typeof args[args.length-1] === 'function') {
      callback = args[args.length-1]
    }
    if (!this.isConnected()) return this.handleError(callback)

    return this.db.collection(collection).insertOne(...args)
  }

  insertMany(collection, ...args) {
    let callback = null
    if(args.length >= 1 && typeof args[args.length-1] === 'function') {
      callback = args[args.length-1]
    }
    if (!this.isConnected()) return this.handleError(callback)

    return this.db.collection(collection).insertMany(...args)
  }

  update(collection, ...args) {
    let callback = null
    if(args.length >= 1 && typeof args[args.length-1] === 'function') {
      callback = args[args.length-1]
    }
    if (!this.isConnected()) return this.handleError(callback)

    return this.db.collection(collection).update(...args)
  }

  updateOne(collection, ...args) {
    let callback = null
    if(args.length >= 1 && typeof args[args.length-1] === 'function') {
      callback = args[args.length-1]
    }
    if (!this.isConnected()) return this.handleError(callback)

    return this.db.collection(collection).updateOne(...args)
  }

  deleteMany(collection, ...args) {
    let callback = null
    if(args.length >= 1 && typeof args[args.length-1] === 'function') {
      callback = args[args.length-1]
    }
    if (!this.isConnected()) return this.handleError(callback)

    return this.db.collection(collection).deleteMany(...args)
  }

  deleteOne(collection, ...args) {
    let callback = null
    if(args.length >= 1 && typeof args[args.length-1] === 'function') {
      callback = args[args.length-1]
    }
    if (!this.isConnected()) return this.handleError(callback)

    return this.db.collection(collection).deleteOne(...args)
  }

  save(collection, ...args) {
    let callback = null
    if(args.length >= 1 && typeof args[args.length-1] === 'function') {
      callback = args[args.length-1]
    }
    if (!this.isConnected()) return this.handleError(callback)

    return this.db.collection(collection).save(...args)
  }

  remove(collection, ...args) {
    let callback = null
    if(args.length >= 1 && typeof args[args.length-1] === 'function') {
      callback = args[args.length-1]
    }
    if (!this.isConnected()) return this.handleError(callback)

    return this.db.collection(collection).remove(...args)
  }

  dropDatabase(...args) {
    return this.db.dropDatabase(...args)
  }

  aggregate(collection, ...args) {
    let callback = null
    if(args.length >= 1 && typeof args[args.length-1] === 'function') {
      callback = args[args.length-1]
    }
    if (!this.isConnected()) return this.handleError(callback)

    return this.db.collection(collection).aggregate(...args)
  }

  queryResource(collection, query, projection, sort, skip, limit) {
    return new Promise((resolve) => {
      if (!this.isConnected()) return this.handleError()
      let queryHash, projectionHash, cur, response;
      if (query != null) {
        try {
          queryHash = JSON.parse(query)
        } catch (err) {
          return this.handleError()
        }
      } else {
        queryHash = {}
      }

      if (projection != null) {
        projectionHash = {};
        let ref1 = projection.split(',');
        for (let j = 0, len = ref1.length; j < len; j++) {
          let p = ref1[j];
          p = p.trim();
          projectionHash[p] = 1;
        }
      }

      cur = this.db.collection(collection).find(queryHash, projectionHash, { batchSize: 50 })
      if (sort != null) {
        let s = JSON.parse(sort);
        sort = {};
        for (let k in s) {
          sort[k] = s[k];
        }
        cur.sort(sort);
      }

      if (skip != null) {
        cur.skip(parseInt(skip));
      }
      if (limit != null) {
        cur.limit(limit = parseInt(limit));
      }
      return cur.count(false, async function (err, total) {
        response = {}
        response['total'] = total
        response['result'] = "[\n"
        let i = 0;
        return cur.each(async function (err, item) {
          if (err) {
            response['result'] = response['result'].concat("\n]");
            return this.handleError()
          }
          if (item != null) {
            if (i++) {
              response['result'] = response['result'].concat(",\n");
            }
            response['result'] = response['result'].concat(JSON.stringify(item));
          }
          if ((item == null) || ((limit != null) && i >= limit)) {
            response['result'] = response['result'].concat("\n]");
            resolve(response)
          }
        })
      })
    })
  }
}

module.exports = MongoClass
