const config = require('./agile-config')

let http = require('http')
const maxBodyLength = 128 * 1024 * 1024

class ProxyHandler {
  constructor(setting, urlStartIndex) {
    if (setting) this.proxy_setting = setting
    else {
      this.proxy_setting = {
        interface: config.get("PROXY_G_INTERFACE"),
        port: config.get("PROXY_G_PORT"),
        ssl: config.get("PROXY_G_SSL")
      }
    }
    // When using proxy to genie nbi from sub routes, use urlStartIndex to intercept correct url.
    if (urlStartIndex) this.urlStartIndex = urlStartIndex
    // since we pass handler to koa, we need to bind this now
    this.handler = this.handler.bind(this)
    this.streamBodyParser = this.streamBodyParser.bind(this)
  }

  // used to parse application/octet-stream to Buffer
  async streamBodyParser(ctx, next) {
    if (ctx.get('content-type').toLowerCase() !== 'application/octet-stream' || ctx.request.rawBody)
      return await next()

    // only parse 'application/octet-stream'
    return new Promise((res, rej) => {
      let temp = [], length = 0
      ctx.req.on('data', trunk => {
        length += trunk.length
        if (length > maxBodyLength) return ctx.throw(413)
        temp.push(trunk)
      })
      ctx.req.on('end', () => res(Buffer.concat(temp)))
      ctx.req.on('error', err => rej(err))
    })
      .then(result => {
        ctx.request.body = result
        ctx.request.rawBody = result
        return next()
      })
      .catch(() => ctx.throw(400))
  }

  async sendRequest(ctx, options, requestData) {
    const headerKeys = ['content-type', 'total', 'cache-control', 'allow']  // use lower case
    return new Promise((resolve, reject) => {
      let clientRequest = http.request(options, res => {
        let responseData = ''
        res.on('data', data => {
          responseData += data
        })
        res.on('end', () => {
          let resHeaderKeys = Object.keys(res.headers)
          ctx.status = res['statusCode']
          headerKeys.forEach(item => {
            let realKey = resHeaderKeys.find(key => key.toLowerCase() === item)
            if (realKey) ctx.set(realKey, res.headers[realKey])
          })
          if (responseData !== '') ctx.body = responseData
          resolve()
        })
      })
      clientRequest.on('error', () => {
        reject(503)
      })
      if (requestData) clientRequest.write(requestData)
      clientRequest.end()
    })
      .catch(err => {
        ctx.throw(err)
      })
  }

  async handler(ctx) {
    if (this.proxy_setting.ssl) return ctx.throw(503)  // no https support now
    let options = {
      method: ctx.method.toUpperCase(),
      host: this.proxy_setting.interface,
      port: this.proxy_setting.port,
      path: ctx.url
    }
    if (this.urlStartIndex) {
      options.path = (ctx.url).substr(this.urlStartIndex)
    }
    let requestData = null
    let reqHeaders = ['Content-Type', 'Content-Length']

    switch (options.method) {
      case 'POST':
      case 'PUT':
        if (ctx.get('fileType') && ctx.get('model') && ctx.get('version')) reqHeaders.push("fileType", "model", "version")
      case 'DELETE':  // eslint-disable-line no-fallthrough
        requestData = ctx.request.rawBody
        reqHeaders.forEach(item => {
          let headerItem = ctx.get(item)
          if (headerItem) {
            if (!options.headers) options.headers = {}
            options.headers[item] = headerItem
          }
        })
      case 'GET':  // eslint-disable-line no-fallthrough
      case "HEAD":
        await this.sendRequest(ctx, options, requestData)
    }
  }
}

module.exports = ProxyHandler
