'use strict'

const crypto = require('crypto')
const dgram = require('dgram')
const http = require('http')
const URL = require('url')

const AgileACSApps = require('tplib/agile-acsapps')
const AgileCacheClass = require('tplib/agile-cache')
const AgileCryptor = require('tplib/agile-cryptor')
const AgileFeatureSet = require('tplib/agile-feature-set')
const MongoClass = require('tplib/mongo-class')
const agileConfig = require('tplib/agile-config')
const stringTable = require('tplib/string-table.json')

const ENCRYPT_ENABLE = agileConfig.get('ENCRYPT_ENABLE')
const dbTplink = new MongoClass(agileConfig.get(stringTable.url.dbTplink))
const dbGenie = new MongoClass(agileConfig.get(stringTable.url.dbGenie))
const agileCache = new AgileCacheClass(agileConfig.getRedisConfig())
const agileCryptor = new AgileCryptor(ENCRYPT_ENABLE, agileConfig.getRedisConfig())
const agileFeatureSet = new AgileFeatureSet()

function parseAuthHeader(authHeader) {
  var method, name, p, part, parts, res, value;
  authHeader = authHeader.trim();
  method = authHeader.split(' ', 1)[0];
  res = {
    method: method
  };
  parts = authHeader.slice(method.length + 1).split(',');
  while ((part = parts.shift()) != null) {
    name = part.split('=', 1)[0];
    if (name.length === part.length) {
      throw new Error('Unable to parse auth header');
    }
    value = part.slice(name.length + 1);
    if (!/^\s*"/.test(value)) {
      value = value.trim();
    } else {
      while (!/[^\\]"\s*$/.test(value)) {
        p = parts.shift();
        if (p == null) {
          throw new Error('Unable to parse auth header');
        }
        value += ',' + p;
      }
      try {
        value = JSON.parse(value);
      } catch (_error) {
        throw new Error('Unable to parse auth header');
      }
    }
    res[name.trim()] = value;
  }
  return res;
}

function basic(username, password) {
  return "Basic " + (new Buffer(username + ":" + password).toString('base64'));
}

function digest(username, password, uri, httpMethod, body, authHeader) {
  var authString, cnonce, ha1, ha2, nc, qop, response;
  cnonce = '0a4f113b';
  nc = '00000001';
  if (authHeader.qop != null) {
    if (authHeader.qop.indexOf(',') !== -1) {
    qop = 'auth';
    } else {
    qop = authHeader.qop;
    }
  }
  ha1 = crypto.createHash('md5');
  ha1.update(username).update(':').update(authHeader.realm).update(':').update(password);
  ha1 = ha1.digest('hex');
  ha2 = crypto.createHash('md5');
  ha2.update(httpMethod).update(':').update(uri);
  if (qop === 'auth-int') {
    ha2.update(':').update(body);
  }
  ha2 = ha2.digest('hex');
  response = crypto.createHash('md5');
  response.update(ha1).update(':').update(authHeader.nonce);
  if (qop != null) {
    response.update(':').update(nc).update(':').update(cnonce).update(':').update(qop);
  }
  response.update(':').update(ha2);
  response = response.digest('hex');
  authString = "Digest username=\"" + username + "\"";
  authString += ",realm=\"" + authHeader.realm + "\"";
  authString += ",nonce=\"" + authHeader.nonce + "\"";
  authString += ",uri=\"" + uri + "\"";
  if (authHeader.algorithm != null) {
    authString += ",algorithm=" + authHeader.algorithm;
  }
  if (qop != null) {
    authString += ",qop=" + qop + ",nc=" + nc + ",cnonce=\"" + cnonce + "\"";
  }
  authString += ",response=\"" + response + "\"";
  if (authHeader.opaque != null) {
    authString += ",opaque=\"" + authHeader.opaque + "\"";
  }
  return authString;
}

function udpConReq(address, un, key, callback) {
  var UDP_CONNECTION_REQUEST_PORT, client, cn, count, f, host, id, message, port, ref, sig, ts, uri;
  if (!address) {
    return false;
  }
  ref = address.split(':', 2), host = ref[0], port = ref[1];
  if (port === void 0) {
    port = 80;
  } else {
    port = parseInt(port);
    if (isNaN(port)) {
    return false;
    }
  }
  ts = Math.trunc(Date.now() / 1000);
  id = Math.trunc(Math.random() * 4294967295);
  cn = crypto.randomBytes(8).toString('hex');
  sig = crypto.createHmac('sha1', key).update("" + ts + id + un + cn).digest('hex');
  uri = "http://" + address + "?ts=" + ts + "&id=" + id + "&un=" + un + "&cn=" + cn + "&sig=" + sig;
  message = Buffer.from("GET " + uri + " HTTP/1.1\r\nHost: " + address + "\r\n\r\n");
  client = dgram.createSocket({
    type: 'udp4',
    reuseAddr: true
  });
  UDP_CONNECTION_REQUEST_PORT = agileConfig.get('UDP_CONNECTION_REQUEST_PORT');
  if (UDP_CONNECTION_REQUEST_PORT) {
    client.bind({
    port: UDP_CONNECTION_REQUEST_PORT,
    exclusive: true
    });
  }
  count = 3;
  client.send(message, 0, message.length, port, host, f = function(err) {
    if (err || --count <= 0) {
    client.close();
    return callback(err);
    }
    return client.send(message, 0, message.length, port, host, f);
  });
  return true;
}

function httpConReq(url, username, password, allowBasicAuth, timeout, callback) {
  var options, request, statusToError;
  options = URL.parse(url);
  if (options.protocol !== 'http:') {
    return callback(new Error('Invalid connection request url or protocol'));
  }
  options.agent = new http.Agent({
    maxSockets: 1
  });
  statusToError = function(statusCode) {
    switch (statusCode) {
    case 200:
    case 204:
      return null;
    case 401:
      return new Error('Incorrect connection request credentials');
    case 0:
      return new Error('Device is unreachable');
    default:
      return new Error("Unexpected response code from device: " + statusCode);
    }
  };
  return request = http.get(options, function(res) {
    var authHeader;
    if (res.statusCode === 401 && (res.headers['www-authenticate'] != null)) {
    authHeader = parseAuthHeader(res.headers['www-authenticate']);
    if (authHeader.method === 'Basic') {
      if (!allowBasicAuth) {
      request.abort();
      callback(new Error('Basic HTTP authentication not allowed'));
      return;
      }
      options.headers = {
      'Authorization': basic(username || '', password || '')
      };
    } else if (authHeader.method === 'Digest') {
      options.headers = {
      'Authorization': digest(username || '', password || '', options.path, 'GET', null, authHeader)
      };
    }
    request = http.get(options, function(res) {
      if (res.statusCode === 0) {
      request = http.get(options, function(res) {
        callback(statusToError(res.statusCode));
        return res.resume();
      }).on('error', function() {
        request.abort();
        return callback(statusToError(0));
      }).on('socket', function(socket) {
        socket.setTimeout(timeout);
        return socket.on('timeout', function() {
        return request.abort();
        });
      });
      } else {
      callback(statusToError(res.statusCode));
      }
      return res.resume();
    }).on('error', function() {
      request.abort();
      return callback(statusToError(0));
    }).on('socket', function(socket) {
      socket.setTimeout(timeout);
      return socket.on('timeout', function() {
      return request.abort();
      });
    });
    } else {
    callback(statusToError(res.statusCode));
    }
    return res.resume();
  }).on('error', function() {
    request.abort();
    return callback(statusToError(0));
  }).on('socket', function(socket) {
    socket.setTimeout(timeout);
    return socket.on('timeout', function() {
    return request.abort();
    });
  });
}

function connectionRequest_dev(deviceId, callback) {
  var CONNECTION_REQUEST_ALLOW_BASIC_AUTH, CONNECTION_REQUEST_TIMEOUT;
  CONNECTION_REQUEST_TIMEOUT = agileConfig.get("CONNECTION_REQUEST_TIMEOUT");
  CONNECTION_REQUEST_ALLOW_BASIC_AUTH = agileConfig.get("CONNECTION_REQUEST_ALLOW_BASIC_AUTH");
  return dbGenie.aggregate(stringTable.collection.devices,
    [
      {
        "$match":{
          "_id": deviceId
        }
      },
      {
        "$project":{
          "ConnReqUrl":{
            "$ifNull":["$Device.ManagementServer.ConnectionRequestURL._value","$InternetGatewayDevice.ManagementServer.ConnectionRequestURL._value"]
          }
          , "UdpConnReqAddr":{
            "$ifNull":["$Device.ManagementServer.UDPConnectionRequestAddress._value","$InternetGatewayDevice.ManagementServer.UDPConnectionRequestAddress._value"]
          }
          , "STUNEnable":{
            "$ifNull":["$Device.ManagementServer.STUNEnable._value","$InternetGatewayDevice.ManagementServer.STUNEnable._value"]
          }
          //, "Username":{
          //	"$ifNull":["$Device.ManagementServer.ConnectionRequestUsername._value","$InternetGatewayDevice.ManagementServer.ConnectionRequestUsername._value"]
          //}
          //, "Password":{
          //	"$ifNull":["$Device.ManagementServer.ConnectionRequestPassword._value","$InternetGatewayDevice.ManagementServer.ConnectionRequestPassword._value"]
          //}
        }
      }
    ]
  , async function (err, results) {
    var conReq, connectionRequestUrl, udpConnectionRequestAddress, stunEnable, username, password;
    if (err) {
    return callback(err);
    }
    if (results.length==0) {
    return callback(new Error("deviceId not found"));
    } else {
    if(ENCRYPT_ENABLE){
      connectionRequestUrl = await agileCryptor.decrypt(results[0].ConnReqUrl);
      udpConnectionRequestAddress = results[0].UdpConnReqAddr != null ? await agileCryptor.decrypt(results[0].UdpConnReqAddr) : void 0;
    } else {
      connectionRequestUrl = results[0].ConnReqUrl;
      udpConnectionRequestAddress = results[0].UdpConnReqAddr != null ? results[0].UdpConnReqAddr : void 0;
    }
    stunEnable = results[0].STUNEnable != null ? results[0].STUNEnable : void 0;
    conReq = function() {
      var udpSent;
      udpSent = udpConReq(udpConnectionRequestAddress, username, password, function(err) {
        if (err) {
        throw err;
        }
      });
      return httpConReq(connectionRequestUrl, username, password, CONNECTION_REQUEST_ALLOW_BASIC_AUTH, CONNECTION_REQUEST_TIMEOUT, function(err) {
        if(stunEnable === 'true'){
          if (udpSent) {
            return callback();
          }
        }
        return callback(err);
      });
    };
    return dbTplink.findOne(stringTable.collection.sysSetting, {
      _id: "cwmp"
    }, async function(err, cwmp){
      if (err || !cwmp || !cwmp.cpe_username || !cwmp.cpe_password) {
        username = deviceId;
        password = "";
      } else {
        if(ENCRYPT_ENABLE){
          username = await agileCryptor.decrypt(cwmp.cpe_username);
          password = await agileCryptor.decrypt(cwmp.cpe_password);
        } else {
          username = cwmp.cpe_username;
          password = cwmp.cpe_password;
        }
      }
      return conReq();
    });
    }
  });
}

async function xmppRequest(deviceId) {
  // find xmpp data
  let xmppData = await dbTplink.findOne(stringTable.collection.xmpp, { "_id":deviceId })
  if(!xmppData) throw new Error(`xmpp data for ${deviceId} not found`)
  if(ENCRYPT_ENABLE) xmppData.xmpp_password = await agileCryptor.decrypt(xmppData.xmpp_password)

  try {
    let connStatus = await AgileACSApps.execCmdWithSafeArgs('/opt/ejabberd-19.05/bin/ejabberdctl', [ 'user_sessions_info', xmppData.username, 'agileacs.com' ])
    if(!connStatus.stdout) throw new Error('')
  } catch (error) {
    throw new Error(`${deviceId} is not connected to xmpp`)
  }

  try {
    await AgileACSApps.execCmdWithSafeArgs('/opt/agileacs/bin/profanityctl', `/connReq ${xmppData.jid} ${xmppData.username} ${xmppData.xmpp_password}`)
  } catch(e) {
    // profanityctl returns 1 even if succeed, check stderr to make sure result
    if(e.stderr) throw new Error(`xmpp command to profanity failed`)
  }

  // add timer here to make sure device is really online
  await agileCache.setPromise(`${deviceId}_xmpp_request_sent`, 1, agileConfig.get('XMPP_CR_TTL'))
  await new Promise((res, rej) => {
    let retries = 0
    let timer = setInterval(async () => {
      try {
        let result = await agileCache.getPromise(`${deviceId}_xmpp_request_sent`)
        if(result !== null) { // not receive device CR yet
          if(retries < agileConfig.get('XMPP_CR_RETRY_TIMES')-1) retries++
          else { // timeout, xmpp cr failed
            agileCache.delPromise(`${deviceId}_xmpp_request_sent`) // no need to await here
            throw new Error('XMPP Connection Request Failed')
          }
        } else {
          clearInterval(timer)
          res()
        }
      } catch(err) {
        clearInterval(timer)
        rej(err)
      }
    }, 500)
  })
}

async function connectionRequest(deviceId, callback) {
  let doXmppRequest = false
   
  try {
    // check if xmpp is enabled
    let sysSetting = await dbTplink.findOne(stringTable.collection.sysSetting, { "_id": "service" })
    if(sysSetting.xmpp !== 'enable') throw new Error('xmpp is not enabled')

    // check if device contains xmpp setting (if device uses xmpp)
    let deviceData = await dbGenie.aggregate(stringTable.collection.devices, [
      {"$match":{"_id": deviceId}},
      {"$project":{"XMPPEnable":{"$ifNull":["$Device.XMPP.Connection.1.Enable._value","$InternetGatewayDevice.XMPP.Connection.1.Enable._value"]}}}
    ]).toArray()
    if (deviceData.length === 0 || !deviceData[0].XMPPEnable ) {
      throw new Error(`${deviceId} is not configured to use xmpp`)
    }

    // check license only when xmpp is pro feature
    if(await agileFeatureSet.isProFeature('xmpp')) {
      // check if pro license enabled
      let isPro = await agileFeatureSet.isProFeatureEnabled()
      if(!isPro) throw new Error('pro feature is not enabled, do not use xmpp')
    }

    // it's xmpp device and xmpp is enabled, try xmpp connection request
    doXmppRequest = true
    await xmppRequest(deviceId)
    callback()
  } catch(err) {
    // console.log(err)
    // check if we need to fallback to CR when xmpp failed
    if(doXmppRequest && !agileConfig.get('XMPP_FAIL_FALLBACK')) {
      return callback(err)
    }
    connectionRequest_dev(deviceId, callback)
  }
}

function connectionRequestPromise(deviceId) {
  return new Promise((res, rej) => {
    connectionRequest(deviceId, err => {
      // console.log(err)
      if(err) return rej(err)
      res()
    })
  })
}

function destructCR() {
  dbTplink.disconnectDb()
  dbGenie.disconnectDb()
  agileCryptor.disconnectRedis()
  agileCache.disconnect()
  agileFeatureSet.destruct()
}

module.exports = {
  destructCR,
  connectionRequest,
  connectionRequestPromise
}
