'use strict'

const koarouter = require('koa-router')

const fs = require("fs")
const os = require("os")

const AgileCacheClass = require('tplib/agile-cache')
const agileConfig = require('tplib/agile-config')

const { saveDiagnosticLock, destoryDiagnosticLock, disconnectCache } = require('tplib/diag_lock_util')
const { connectionRequestPromise } = require('../connection_request/connection_request')
const { getDiagnosticTimeout, insertTasks, watchTask, qureyTask, taskLibDestruct, checkTasksFaults } = require('./tasks_lib')

const subroute = new koarouter()
const subroutepath = __filename.split('/').pop().split('.')[0]
const agileCache = new AgileCacheClass(agileConfig.getRedisConfig())

subroute.use(require('koa-bodyparser')())

subroute.get(`/:id/${subroutepath}`, async ctx => {
  let deviceId = ctx.params['id']
  try {
    let result = await checkTasksFaults(deviceId)
    ctx.body = result
  } catch (err) {
    console.log(err)
    return ctx.throw(400)
  }
})

// device tasks
subroute.post(`/:id/${subroutepath}`, async ctx => {
  let deviceId = ctx.params['id']
  if (Object.keys(ctx.request.body).length !== 0) {
    let task = ctx.request.body
    task.device = deviceId

    // seems diag actions doesn't effects http response, no need to promisify
    if (task.diagnostic && task.diagnostic !== 'UDPEchoConfig') {
      if (task.cancel) {
        console.log(`device tasks destoryDiagnosticLock: ${deviceId}`)
        destoryDiagnosticLock(deviceId, result => {
          if (result.result === 'error') console.log("Error: destroyDiagnosticLock failed.")
        })
      } else {
        let timeout = getDiagnosticTimeout(task)
        saveDiagnosticLock(deviceId, task.diagnostic, Math.floor(timeout / 1000), result => {
          if (result.result === 'error') console.log('Error: saveDiagnosticLock failed.')
        })
      }
    }

    if (task.name == 'isOnline') {
      try {
        await connectionRequestPromise(deviceId)
        return response2xx(ctx, 200, null, task)
      } catch (err) {
        // console.log(err)
        // log with reason if device is unreachable
        let timestamp = new Date().toIsoString()
        let message = `${timestamp} [INFO] 6 127.0.0.1 ${deviceId}: ${err.message}${os.EOL}`
        fs.appendFile('/var/log/genieacs/cwmp-access.log', message, err => {
          console.log(err)
        })
        return response2xx(ctx, 202, err.message, task)
      }
    }

    // task.name !== 'isOnline'
    try {
      await insertTasks(task)
      await agileCache.delPromise(`${deviceId}_tasks_faults_operations`)
    } catch (err) {
      return ctx.throw(400)
    }

    if (ctx.query.connection_request == null) return response2xx(ctx, 202, null, task)

    // ctx.query.connection_request !== null
    try {
      await connectionRequestPromise(deviceId)
    } catch (err) {
      console.log(err)
      return response2xx(ctx, 202, err.message, task)
    }

    if ((task.name == 'getParameterValues' || task.name == 'refreshObject') && ctx.query.device_online_threshold == null) {
      try {
        let { status, fault } = await watchTask(deviceId, task._id, agileConfig.get('DEVICE_ONLINE_THRESHOLD_GENERAL'))
        if (status === 'timeout') return response2xx(ctx, 202, '[Task timeout] Task queued but not processed', task)
        if (status === 'fault') {
          if (!fault) throw new Error('no session context fault string')
          let dbTask = await qureyTask(task._id)
          let resBody = ctx.body = {
            task_id: dbTask._id, fault_id: fault._id,
            fault_code: fault.code, fault_message: fault.message, fault: fault
          }
          return response2xx(ctx, 202, 'Task faulted', resBody)
        }
        return response2xx(ctx, 200, null, task)
      } catch (err) {
        console.log(err)
        return ctx.throw(400)
      }
    } else {
      try {
        let { status, fault } = await watchTask(deviceId, task._id, agileConfig.get('DEVICE_ONLINE_THRESHOLD'))
        if (status === 'timeout') return response2xx(ctx, 202, '[Task timeout] The CPE does not reply within the required time, please refresh to check whether the settings have taken effect.', task)
        if (status === 'fault') {
          if (!fault) throw new Error('no session context fault string')
          let resBody = {
            task_id: task._id, fault_id: `${deviceId}:task_${task._id}`,
            fault_code: fault.code, fault_message: fault.message, fault: fault
          }
          if (task.diagnostic) {
            destoryDiagnosticLock(deviceId, result => {
              if (result.result === 'error') console.log("Error: destroyDiagnosticLock failed when fault.")
            })
          }
          return response2xx(ctx, 202, 'Task faulted', resBody)
        }
        return response2xx(ctx, 200, null, task)
      } catch (err) {
        console.log(err)
        return ctx.throw(400)
      }
    }
  } else if (ctx.query.connection_request !== null) {
    try {
      await connectionRequestPromise(deviceId)
      ctx.status = 200
    } catch (err) {
      console.log(err)
      Object.assign(ctx, { status: 504, body: `${err.name}: ${err.message}` })
    }
  } else {
    ctx.throw(400)
  }
})

subroute.delete(`/:id/${subroutepath}`, ctx => {
  let task = ctx.request.body
  task.device = ctx.params['id']
  destoryDiagnosticLock(ctx.params['id'], result => {
    if (result.result == 'error') console.log("Error: destroyDiagnosticLock failed.")
  })
  return response2xx(ctx, 200, null, task)
})

function response2xx(ctx, status, message, body) {
  Object.assign(ctx, { status, body })
  if (message && message !== '') ctx.res.statusMessage = message
}

subroute.destruct = () => {
  agileCache.disconnect()
  disconnectCache()
  taskLibDestruct()
}

module.exports = subroute
