'use strict'

const AgileCacheClass = require('tplib/agile-cache')
const MongoClass = require('tplib/mongo-class')
const agileConfig = require('tplib/agile-config')
const stringTable = require('tplib/string-table.json')

const dbGenie = new MongoClass(agileConfig.get(stringTable.url.dbGenie))
const agileCache = new AgileCacheClass(agileConfig.getRedisConfig())

async function qureyTask(_id, options) {
  return dbGenie.findOne(stringTable.collection.tasks, { _id }, options)
}

async function checkTasksFaults(deviceId) {
  return new Promise((resolve, reject) => {
    let result = {}
    dbGenie.find(stringTable.collection.tasks, { device: deviceId })
      .then(tasks => {
        result['total'] = tasks.length
        result['result'] = tasks
        if (result['total'] > 0) {
          let taskId = tasks[0]._id
          return dbGenie.findOne(stringTable.collection.faults, {
            _id: deviceId + ":task_" + taskId
          }).then(fault => {
            if (fault) {
              if (fault['provisions']) fault['provisions'] = JSON.parse(fault['provisions'])
              result["fault"] = fault
            }
            resolve(result)
          }).catch(err => {
            reject(err)
          })
        } else {
          resolve(result)
        }
      })
      .catch(err => {
        reject(err)
      })
  })
}

function getDiagnosticTimeout(task) {
  var MaxHopCount, NumberOfRepetitions, NumberOfTries, Timeout, UploadFileSize, UploadTime, timeout
  timeout = 10000
  switch (task.diagnostic) {
    case 'IPPingDiagnostics':
      NumberOfRepetitions = 3
      Timeout = 5000
      task.parameterValues.forEach(function (k) {
        if (k[0].indexOf('NumberOfRepetitions') >= 0) {
          NumberOfRepetitions = k[1] - 0
        }
        if (k[0].indexOf('Timeout') >= 0) {
          Timeout = k[1] - 0
        }
      })
      timeout = NumberOfRepetitions * Timeout + 5 * 1000
      break
    case 'TraceRouteDiagnostics':
      NumberOfTries = 3
      MaxHopCount = 30
      Timeout = 5000
      task.parameterValues.forEach(function (k) {
        if (k[0].indexOf('NumberOfTries') >= 0) {
          NumberOfTries = k[1] - 0
        }
        if (k[0].indexOf('MaxHopCount') >= 0) {
          MaxHopCount = k[1] - 0
        }
        if (k[0].indexOf('Timeout') >= 0) {
          Timeout = k[1] - 0
        }
      })
      timeout = NumberOfTries * MaxHopCount * (Timeout + 1000) + 5 * 1000
      break
    case 'UploadDiagnostics':
      UploadFileSize = 0
      task.parameterValues.forEach(function (k) {
        if (k[0].indexOf('TestFileLength') >= 0) {
          UploadFileSize = k[1] - 0
        }
      })
      UploadTime = (UploadFileSize - 0) / 100000
      timeout = (20 + Math.ceil(UploadTime)) * 1000
      break
    case 'DownloadDiagnostics':
      timeout = 1020 * 1000
      break
    case 'UDPEchoConfig':
      timeout = 300 * 1000
      break
    case 'NSLookupDiagnostics':
      timeout = 10 * 1000
      break
    default:
      timeout = 10 * 1000
      break
  }
  return timeout
}

function sanitizeTask(task) {
  task.timestamp = new Date(task.timestamp || Date.now())
  if (task.expiry != null) {
    if (Object.prototype.toString.call(task.expiry) === '[object Date]' || isNaN(task.expiry)) {
      task.expiry = new Date(task.expiry)
    } else {
      task.expiry = new Date(task.timestamp.getTime() + task.expiry * 1000)
    }
  }
}

async function insertTasks(tasks) {
  if ((tasks != null) && Object.prototype.toString.call(tasks) !== '[object Array]') {
    tasks = [tasks]
  } else if ((tasks == null) || tasks.length === 0) return (tasks || [])

  return Promise.all(tasks.map(task => {
    sanitizeTask(task)
    if (task.uniqueKey != null) {
      return dbGenie.remove(stringTable.collection.tasks, { device: task.device, uniqueKey: task.uniqueKey })
    }
  }))
    .then(() => {
      return dbGenie.insert(stringTable.collection.tasks, tasks)
    })
}

async function watchTask(deviceId, taskId, timeout) {
  const checkPeriod = 500
  let findTask = (res, rej) => {
    qureyTask(taskId, { '_id': 1 })
      .then(task => {
        if (!task) return res({ status: 'completed' })
        // check if there is fault
        return dbGenie.findOne(stringTable.collection.faults, { 
          _id: deviceId + ":task_" + taskId  
        }, {
          '_id': 1,
          'code': 1,
          'message': 1,
          'detail': 1
        }).then(fault => {
          if (fault) return res({ status: 'fault', fault })
          if ((timeout -= checkPeriod) <= 0) return res({ status: 'timeout' })
          // check again after checkPeriod
          setTimeout(() => {
            findTask(res, rej)
          }, checkPeriod)
        })
        .catch(err => {
          rej(err)
        })
      })
      .catch(err => {
        rej(err)
      })
  }

  return new Promise((res, rej) => {
    findTask(res, rej)
  })
}

function taskLibDestruct() {
  dbGenie.disconnectDb()
  agileCache.disconnect()
}

module.exports = {
  insertTasks,
  getDiagnosticTimeout,
  qureyTask,
  taskLibDestruct,
  watchTask,
  checkTasksFaults
}
